# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

from torch import nn

from zeta.nn.embeddings.base import BaseEmbedding

# Other embedding


class NominalEmbedding(BaseEmbedding):
    def forward(self, num_tokens: int, dim: int) -> nn.Module:
        embedding = nn.Embedding(num_tokens, dim)

        return embedding



import functools
import math
from operator import mul

import torch
import torch.nn.functional as F
from einops import pack, rearrange, repeat, unpack
from einops.layers.torch import Rearrange
from torch import nn

from zeta.nn.attention.attend import Attend

# helper functions


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def mul_reduce(tup):
    return functools.reduce(mul, tup)


def divisible_by(numer, denom):
    return (numer % denom) == 0


mlist = nn.ModuleList

# for time conditioning


class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim, theta=10000):
        super().__init__()
        self.theta = theta
        self.dim = dim

    def forward(self, x):
        dtype, device = x.dtype, x.device
        assert (
            dtype == torch.float
        ), "input to sinusoidal pos emb must be a float type"

        half_dim = self.dim // 2
        emb = math.log(self.theta) / (half_dim - 1)
        emb = torch.exp(
            torch.arange(half_dim, device=device, dtype=dtype) * -emb
        )
        emb = rearrange(x, "i -> i 1") * rearrange(emb, "j -> 1 j")
        return torch.cat((emb.sin(), emb.cos()), dim=-1).type(dtype)


# layernorm 3d


class RMSNorm(nn.Module):
    def __init__(self, chan, dim=1):
        super().__init__()
        self.dim = dim
        self.gamma = nn.Parameter(torch.ones(chan))

    def forward(self, x):
        dim = self.dim
        right_ones = (dim + 1) if dim < 0 else (x.ndim - 1 - dim)
        gamma = self.gamma.reshape(-1, *((1,) * right_ones))
        return F.normalize(x, dim=dim) * (x.shape[dim] ** 0.5) * gamma


# FeedForwardV


def shift_token(t):
    t, t_shift = t.chunk(2, dim=1)
    t_shift = F.pad(t_shift, (0, 0, 0, 0, 1, -1), value=0.0)
    return torch.cat((t, t_shift), dim=1)


class GEGLU(nn.Module):
    def forward(self, x):
        x, gate = x.chunk(2, dim=1)
        return x * F.gelu(gate)


class FeedForwardV(nn.Module):
    def __init__(self, dim, mult=4):
        super().__init__()

        inner_dim = int(dim * mult * 2 / 3)
        self.proj_in = nn.Sequential(
            nn.Conv3d(dim, inner_dim * 2, 1, bias=False), GEGLU()
        )

        self.proj_out = nn.Sequential(
            RMSNorm(inner_dim), nn.Conv3d(inner_dim, dim, 1, bias=False)
        )

    def forward(self, x, enable_time=True):
        is_video = x.ndim == 5
        enable_time &= is_video

        if not is_video:
            x = rearrange(x, "b c h w -> b c 1 h w")

        x = self.proj_in(x)

        if enable_time:
            x = shift_token(x)

        out = self.proj_out(x)

        if not is_video:
            out = rearrange(out, "b c 1 h w -> b c h w")

        return out


# best relative positional encoding


class ContinuousPositionBias(nn.Module):
    """from https://arxiv.org/abs/2111.09883"""

    def __init__(self, *, dim, heads, num_dims=1, layers=2):
        super().__init__()
        self.num_dims = num_dims

        self.net = nn.ModuleList([])
        self.net.append(nn.Sequential(nn.Linear(self.num_dims, dim), nn.SiLU()))

        for _ in range(layers - 1):
            self.net.append(nn.Sequential(nn.Linear(dim, dim), nn.SiLU()))

        self.net.append(nn.Linear(dim, heads))

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, *dimensions):
        device = self.device

        shape = torch.tensor(dimensions, device=device)
        rel_pos_shape = 2 * shape - 1

        # calculate strides

        strides = torch.flip(rel_pos_shape, (0,)).cumprod(dim=-1)
        strides = torch.flip(F.pad(strides, (1, -1), value=1), (0,))

        # get all positions and calculate all the relative distances

        positions = [torch.arange(d, device=device) for d in dimensions]
        grid = torch.stack(torch.meshgrid(*positions, indexing="ij"), dim=-1)
        grid = rearrange(grid, "... c -> (...) c")
        rel_dist = rearrange(grid, "i c -> i 1 c") - rearrange(
            grid, "j c -> 1 j c"
        )

        # get all relative positions across all dimensions

        rel_positions = [
            torch.arange(-d + 1, d, device=device) for d in dimensions
        ]
        rel_pos_grid = torch.stack(
            torch.meshgrid(*rel_positions, indexing="ij"), dim=-1
        )
        rel_pos_grid = rearrange(rel_pos_grid, "... c -> (...) c")

        # mlp input

        bias = rel_pos_grid.float()

        for layer in self.net:
            bias = layer(bias)

        # convert relative distances to indices of the bias

        rel_dist += shape - 1  # make sure all positive
        rel_dist *= strides
        rel_dist_indices = rel_dist.sum(dim=-1)

        # now select the bias for each unique relative position combination

        bias = bias[rel_dist_indices]
        return rearrange(bias, "i j h -> h i j")


# helper classes


class Attention(nn.Module):
    def __init__(self, dim, dim_head=64, heads=8, flash=False, causal=False):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = dim_head * heads

        self.attend = Attend(flash=flash, causal=causal)

        self.norm = RMSNorm(dim, dim=-1)

        self.to_q = nn.Linear(dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(dim, inner_dim * 2, bias=False)
        self.to_out = nn.Linear(inner_dim, dim, bias=False)

        nn.init.zeros_(self.to_out.weight.data)  # identity with skip connection

    def forward(self, x, rel_pos_bias=None):
        x = self.norm(x)

        q, k, v = self.to_q(x), *self.to_kv(x).chunk(2, dim=-1)

        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads),
            (q, k, v),
        )

        out = self.attend(q, k, v, bias=rel_pos_bias)

        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


# main contribution - pseudo 3d conv


class PseudoConv3d(nn.Module):
    def __init__(
        self,
        dim,
        dim_out=None,
        kernel_size=3,
        *,
        temporal_kernel_size=None,
        **kwargs,
    ):
        super().__init__()
        dim_out = default(dim_out, dim)
        temporal_kernel_size = default(temporal_kernel_size, kernel_size)

        self.spatial_conv = nn.Conv2d(
            dim, dim_out, kernel_size=kernel_size, padding=kernel_size // 2
        )

        self.temporal_conv = (
            nn.Conv1d(
                dim_out,
                dim_out,
                kernel_size=temporal_kernel_size,
                padding=temporal_kernel_size // 2,
            )
            if kernel_size > 1
            else None
        )

        if exists(self.temporal_conv):
            nn.init.dirac_(
                self.temporal_conv.weight.data
            )  # initialized to be identity
            nn.init.zeros_(self.temporal_conv.bias.data)

    def forward(self, x, enable_time=True):
        b, c, *_, h, w = x.shape

        is_video = x.ndim == 5
        enable_time &= is_video

        if is_video:
            x = rearrange(x, "b c f h w -> (b f) c h w")

        x = self.spatial_conv(x)

        if is_video:
            x = rearrange(x, "(b f) c h w -> b c f h w", b=b)

        if not enable_time or not exists(self.temporal_conv):
            return x

        x = rearrange(x, "b c f h w -> (b h w) c f")

        x = self.temporal_conv(x)

        x = rearrange(x, "(b h w) c f -> b c f h w", h=h, w=w)

        return x


# factorized spatial temporal attention from Ho et al.


class SpatioTemporalAttention(nn.Module):
    def __init__(
        self,
        dim,
        *,
        dim_head=64,
        heads=8,
        add_feed_forward=True,
        ff_mult=4,
        pos_bias=True,
        flash=False,
        causal_time_attn=False,
    ):
        super().__init__()
        assert not (flash and pos_bias), (
            "learned positional attention bias is not compatible with flash"
            " attention"
        )

        self.spatial_attn = Attention(
            dim=dim, dim_head=dim_head, heads=heads, flash=flash
        )

        self.spatial_rel_pos_bias = (
            ContinuousPositionBias(dim=dim // 2, heads=heads, num_dims=2)
            if pos_bias
            else None
        )

        self.temporal_attn = Attention(
            dim=dim,
            dim_head=dim_head,
            heads=heads,
            flash=flash,
            causal=causal_time_attn,
        )

        self.temporal_rel_pos_bias = (
            ContinuousPositionBias(dim=dim // 2, heads=heads, num_dims=1)
            if pos_bias
            else None
        )

        self.has_feed_forward = add_feed_forward
        if not add_feed_forward:
            return

        self.ff = FeedForwardV(dim=dim, mult=ff_mult)

    def forward(self, x, enable_time=True):
        b, c, *_, h, w = x.shape
        is_video = x.ndim == 5
        enable_time &= is_video

        if is_video:
            x = rearrange(x, "b c f h w -> (b f) (h w) c")
        else:
            x = rearrange(x, "b c h w -> b (h w) c")

        space_rel_pos_bias = (
            self.spatial_rel_pos_bias(h, w)
            if exists(self.spatial_rel_pos_bias)
            else None
        )

        x = self.spatial_attn(x, rel_pos_bias=space_rel_pos_bias) + x

        if is_video:
            x = rearrange(x, "(b f) (h w) c -> b c f h w", b=b, h=h, w=w)
        else:
            x = rearrange(x, "b (h w) c -> b c h w", h=h, w=w)

        if enable_time:
            x = rearrange(x, "b c f h w -> (b h w) f c")

            time_rel_pos_bias = (
                self.temporal_rel_pos_bias(x.shape[1])
                if exists(self.temporal_rel_pos_bias)
                else None
            )

            x = self.temporal_attn(x, rel_pos_bias=time_rel_pos_bias) + x

            x = rearrange(x, "(b h w) f c -> b c f h w", w=w, h=h)

        if self.has_feed_forward:
            x = self.ff(x, enable_time=enable_time) + x

        return x


# resnet block
class Block(nn.Module):
    def __init__(
        self, dim, dim_out, kernel_size=3, temporal_kernel_size=None, groups=8
    ):
        super().__init__()
        self.project = PseudoConv3d(dim, dim_out, 3)
        self.norm = nn.GroupNorm(groups, dim_out)
        self.act = nn.SiLU()

    def forward(self, x, scale_shift=None, enable_time=False):
        x = self.project(x, enable_time=enable_time)
        x = self.norm(x)

        if exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1) + shift

        return self.act(x)


class ResnetBlock(nn.Module):
    def __init__(self, dim, dim_out, *, timestep_cond_dim=None, groups=8):
        super().__init__()

        self.timestep_mlp = None

        if exists(timestep_cond_dim):
            self.timestep_mlp = nn.Sequential(
                nn.SiLU(), nn.Linear(timestep_cond_dim, dim_out * 2)
            )

        self.block1 = Block(dim, dim_out, groups=groups)
        self.block2 = Block(dim_out, dim_out, groups=groups)
        self.res_conv = (
            PseudoConv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()
        )

    def forward(self, x, timestep_emb=None, enable_time=True):
        assert not (exists(timestep_emb) ^ exists(self.timestep_mlp))

        scale_shift = None

        if exists(self.timestep_mlp) and exists(timestep_emb):
            time_emb = self.timestep_mlp(timestep_emb)
            to_einsum_eq = "b c 1 1 1" if x.ndim == 5 else "b c 1 1"
            time_emb = rearrange(time_emb, f"b c -> {to_einsum_eq}")
            scale_shift = time_emb.chunk(2, dim=1)

        h = self.block1(x, scale_shift=scale_shift, enable_time=enable_time)

        h = self.block2(h, enable_time=enable_time)

        return h + self.res_conv(x)


# pixelshuffle upsamples and downsamples
# where time dimension can be configured
class Downsample(nn.Module):
    def __init__(
        self,
        dim: int,
        downsample_space: bool = True,
        downsample_time=False,
        nonlin=False,
    ):
        super().__init__()
        assert downsample_space or downsample_time

        self.down_space = (
            nn.Sequential(
                Rearrange("b c (h p1) (w p2) -> b (c p1 p2) h w", p1=2, p2=2),
                nn.Conv2d(dim * 4, dim, 1, bias=False),
                nn.SiLU() if nonlin else nn.Identity(),
            )
            if downsample_space
            else None
        )

        self.down_time = (
            nn.Sequential(
                Rearrange("b c (f p) h w -> b (c p) f h w", p=2),
                nn.Conv3d(dim * 2, dim, 1, bias=False),
                nn.SiLU() if nonlin else nn.Identity(),
            )
            if downsample_time
            else None
        )

    def forward(self, x, enable_time=True):
        is_video = x.ndim == 5

        if is_video:
            x = rearrange(x, "b c f h w -> b f c h w")
            x, ps = pack([x], "* c h w")

        if exists(self.down_space):
            x = self.down_space(x)

        if is_video:
            (x,) = unpack(x, ps, "* c h w")
            x = rearrange(x, "b f c h w -> b c f h w")

        if not is_video or not exists(self.down_time) or not enable_time:
            return x

        x = self.down_time(x)

        return x


class Upsample(nn.Module):
    def __init__(
        self, dim, upsample_space=True, upsample_time=False, nonlin=False
    ):
        super().__init__()
        assert upsample_space or upsample_time

        self.up_space = (
            nn.Sequential(
                nn.Conv2d(dim, dim * 4, 1),
                nn.SiLU() if nonlin else nn.Identity(),
                Rearrange("b (c p1 p2) h w -> b c (h p1) (w p2)", p1=2, p2=2),
            )
            if upsample_space
            else None
        )

        self.up_time = (
            nn.Sequential(
                nn.Conv3d(dim, dim * 2, 1),
                nn.SiLU() if nonlin else nn.Identity(),
                Rearrange("b (c p) f h w -> b c (f p) h w", p=2),
            )
            if upsample_time
            else None
        )

        self.init_()

    def init_(self):
        if exists(self.up_space):
            self.init_conv_(self.up_space[0], 4)

        if exists(self.up_time):
            self.init_conv_(self.up_time[0], 2)

    def init_conv_(self, conv, factor):
        o, *remain_dims = conv.weight.shape
        conv_weight = torch.empty(o // factor, *remain_dims)
        nn.init.kaiming_uniform_(conv_weight)
        conv_weight = repeat(conv_weight, "o ... -> (o r) ...", r=factor)

        conv.weight.data.copy_(conv_weight)
        nn.init.zeros_(conv.bias.data)

    def forward(self, x, enable_time=True):
        is_video = x.ndim == 5

        if is_video:
            x = rearrange(x, "b c f h w -> b f c h w")
            x, ps = pack([x], "* c h w")

        if exists(self.up_space):
            x = self.up_space(x)

        if is_video:
            (x,) = unpack(x, ps, "* c h w")
            x = rearrange(x, "b f c h w -> b c f h w")

        if not is_video or not exists(self.up_time) or not enable_time:
            return x

        x = self.up_time(x)

        return x


# space time factorized 3d unet
class SpaceTimeUnet(nn.Module):
    def __init__(
        self,
        *,
        dim,
        channels=3,
        dim_mult=(1, 2, 4, 8),
        self_attns=(False, False, False, True),
        temporal_compression=(False, True, True, True),
        resnet_block_depths=(2, 2, 2, 2),
        attn_dim_head=64,
        attn_heads=8,
        condition_on_timestep=True,
        attn_pos_bias=True,
        flash_attn=False,
        causal_time_attn=False,
    ):
        super().__init__()
        assert (
            len(dim_mult)
            == len(self_attns)
            == len(temporal_compression)
            == len(resnet_block_depths)
        )

        num_layers = len(dim_mult)

        dims = [dim, *map(lambda mult: mult * dim, dim_mult)]
        dim_in_out = zip(dims[:-1], dims[1:])

        # determine the valid multiples of the image size and frames of the video

        self.frame_multiple = 2 ** sum(tuple(map(int, temporal_compression)))
        self.image_size_multiple = 2**num_layers

        # timestep conditioning for DDPM, not to be confused with the time dimension of the video

        self.to_timestep_cond = None
        timestep_cond_dim = (dim * 4) if condition_on_timestep else None

        if condition_on_timestep:
            self.to_timestep_cond = nn.Sequential(
                SinusoidalPosEmb(dim),
                nn.Linear(dim, timestep_cond_dim),
                nn.SiLU(),
            )

        # layers

        self.downs = mlist([])
        self.ups = mlist([])

        attn_kwargs = dict(
            dim_head=attn_dim_head,
            heads=attn_heads,
            pos_bias=attn_pos_bias,
            flash=flash_attn,
            causal_time_attn=causal_time_attn,
        )

        mid_dim = dims[-1]

        self.mid_block1 = ResnetBlock(
            mid_dim, mid_dim, timestep_cond_dim=timestep_cond_dim
        )

        self.mid_attn = SpatioTemporalAttention(dim=mid_dim, **attn_kwargs)
        self.mid_block2 = ResnetBlock(
            mid_dim, mid_dim, timestep_cond_dim=timestep_cond_dim
        )

        for (
            _,
            self_attend,
            (dim_in, dim_out),
            compress_time,
            resnet_block_depth,
        ) in zip(
            range(num_layers),
            self_attns,
            dim_in_out,
            temporal_compression,
            resnet_block_depths,
        ):
            assert resnet_block_depth >= 1

            self.downs.append(
                mlist(
                    [
                        ResnetBlock(
                            dim_in, dim_out, timestep_cond_dim=timestep_cond_dim
                        ),
                        mlist(
                            [
                                ResnetBlock(dim_out, dim_out)
                                for _ in range(resnet_block_depth)
                            ]
                        ),
                        (
                            SpatioTemporalAttention(dim=dim_out, **attn_kwargs)
                            if self_attend
                            else None
                        ),
                        Downsample(dim_out, downsample_time=compress_time),
                    ]
                )
            )

            self.ups.append(
                mlist(
                    [
                        ResnetBlock(
                            dim_out * 2,
                            dim_in,
                            timestep_cond_dim=timestep_cond_dim,
                        ),
                        mlist(
                            [
                                ResnetBlock(
                                    dim_in + (dim_out if ind == 0 else 0),
                                    dim_in,
                                )
                                for ind in range(resnet_block_depth)
                            ]
                        ),
                        (
                            SpatioTemporalAttention(dim=dim_in, **attn_kwargs)
                            if self_attend
                            else None
                        ),
                        Upsample(dim_out, upsample_time=compress_time),
                    ]
                )
            )

        self.skip_scale = 2**-0.5  # paper shows faster convergence

        self.conv_in = PseudoConv3d(
            dim=channels, dim_out=dim, kernel_size=7, temporal_kernel_size=3
        )

        self.conv_out = PseudoConv3d(
            dim=dim, dim_out=channels, kernel_size=3, temporal_kernel_size=3
        )

    def forward(self, x, timestep=None, enable_time=True):
        # some asserts

        assert not (exists(self.to_timestep_cond) ^ exists(timestep))
        is_video = x.ndim == 5

        if enable_time and is_video:
            frames = x.shape[2]
            assert divisible_by(frames, self.frame_multiple), (
                f"number of frames on the video ({frames}) must be divisible by"
                f" the frame multiple ({self.frame_multiple})"
            )

        height, width = x.shape[-2:]
        assert divisible_by(height, self.image_size_multiple) and divisible_by(
            width, self.image_size_multiple
        ), (
            "height and width of the image or video must be a multiple of"
            f" {self.image_size_multiple}"
        )

        # main logic

        t = (
            self.to_timestep_cond(rearrange(timestep, "... -> (...)"))
            if exists(timestep)
            else None
        )

        x = self.conv_in(x, enable_time=enable_time)

        hiddens = []

        for init_block, blocks, maybe_attention, downsample in self.downs:
            x = init_block(x, t, enable_time=enable_time)

            hiddens.append(x.clone())

            for block in blocks:
                x = block(x, enable_time=enable_time)

            if exists(maybe_attention):
                x = maybe_attention(x, enable_time=enable_time)

            hiddens.append(x.clone())

            x = downsample(x, enable_time=enable_time)

        x = self.mid_block1(x, t, enable_time=enable_time)
        x = self.mid_attn(x, enable_time=enable_time)
        x = self.mid_block2(x, t, enable_time=enable_time)

        for init_block, blocks, maybe_attention, upsample in reversed(self.ups):
            x = upsample(x, enable_time=enable_time)

            x = torch.cat((hiddens.pop() * self.skip_scale, x), dim=1)

            x = init_block(x, t, enable_time=enable_time)

            x = torch.cat((hiddens.pop() * self.skip_scale, x), dim=1)

            for block in blocks:
                x = block(x, enable_time=enable_time)

            if exists(maybe_attention):
                x = maybe_attention(x, enable_time=enable_time)

        x = self.conv_out(x, enable_time=enable_time)
        return x



"""Kolmogorov-Arnold Transformer (KAT) model
author = "Xingyi Yang"
"""

import math
from functools import partial

import torch
import torch.nn as nn
import torch.utils.checkpoint
from kat_rational import KAT_Group
from timm.models.layers import to_2tuple

# git clone https://github.com/Adamdad/rational_kat_cu.git
# cd rational_kat_cu
# pip install -e .


def calculate_gain(nonlinearity, param=None):
    r"""Return the recommended gain value for the given nonlinearity function.
    The values are as follows:

    ================= ====================================================
    nonlinearity      gain
    ================= ====================================================
    Linear / Identity :math:`1`
    Conv{1,2,3}D      :math:`1`
    Sigmoid           :math:`1`
    Tanh              :math:`\frac{5}{3}`
    ReLU              :math:`\sqrt{2}`
    Leaky Relu        :math:`\sqrt{\frac{2}{1 + \text{negative\_slope}^2}}`
    SELU              :math:`\frac{3}{4}`
    ================= ====================================================

    .. warning::
        In order to implement `Self-Normalizing Neural Networks`_ ,
        you should use ``nonlinearity='linear'`` instead of ``nonlinearity='selu'``.
        This gives the initial weights a variance of ``1 / N``,
        which is necessary to induce a stable fixed point in the forward pass.
        In contrast, the default gain for ``SELU`` sacrifices the normalisation
        effect for more stable gradient flow in rectangular layers.

    Args:
        nonlinearity: the non-linear function (`nn.functional` name)
        param: optional parameter for the non-linear function

    Examples:
        >>> gain = nn.init.calculate_gain('leaky_relu', 0.2)  # leaky_relu with negative_slope=0.2

    .. _Self-Normalizing Neural Networks: https://papers.nips.cc/paper/2017/hash/5d44ee6f2c3f71b73125876103c8f6c4-Abstract.html
    """
    linear_fns = [
        "linear",
        "conv1d",
        "conv2d",
        "conv3d",
        "conv_transpose1d",
        "conv_transpose2d",
        "conv_transpose3d",
    ]
    if nonlinearity in linear_fns or nonlinearity == "sigmoid":
        return 1
    elif nonlinearity == "tanh":
        return 5.0 / 3
    elif nonlinearity == "relu":
        return math.sqrt(2.0)
    elif nonlinearity == "gelu":
        return math.sqrt(2.3567850379928976)
    elif nonlinearity == "silu" or nonlinearity == "swish":
        return math.sqrt(2.8178205270359653)
    elif nonlinearity == "leaky_relu":
        if param is None:
            negative_slope = 0.01
        elif (
            not isinstance(param, bool)
            and isinstance(param, int)
            or isinstance(param, float)
        ):
            # True/False are instances of int, hence check above
            negative_slope = param
        else:
            raise ValueError(
                "negative_slope {} not a valid number".format(param)
            )
        return math.sqrt(2.0 / (1 + negative_slope**2))
    elif nonlinearity == "selu":
        return (
            3.0 / 4
        )  # Value found empirically (https://github.com/pytorch/pytorch/pull/50664)
    else:
        raise ValueError("Unsupported nonlinearity {}".format(nonlinearity))


torch.nn.init.calculate_gain = calculate_gain


class KAN(nn.Module):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=KAT_Group,
        norm_layer=None,
        bias=True,
        drop=0.0,
        use_conv=False,
        act_init="gelu",
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)
        linear_layer = (
            partial(nn.Conv2d, kernel_size=1) if use_conv else nn.Linear
        )

        self.fc1 = linear_layer(in_features, hidden_features, bias=bias[0])
        self.act1 = KAT_Group(mode="identity")
        self.drop1 = nn.Dropout(drop_probs[0])
        self.norm = (
            norm_layer(hidden_features)
            if norm_layer is not None
            else nn.Identity()
        )
        self.act2 = KAT_Group(mode=act_init)
        self.fc2 = linear_layer(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x):
        x = self.act1(x)
        x = self.drop1(x)
        x = self.fc1(x)
        x = self.act2(x)
        x = self.drop2(x)
        x = self.fc2(x)
        return x



import torch
from torch import nn


class Lora(nn.Module):
    """
    Lora module applies a linear transformation to the input tensor using the Lora algorithm.

    Args:
        dim (int): The input dimension.
        dim_out (int): The output dimension.
        r (int, optional): The rank of the transformation. Defaults to 8.
        alpha (float, optional): The scaling factor. Defaults to None.

    Attributes:
        scale (float): The scaling factor calculated as alpha / r.
        A (nn.Parameter): The learnable parameter representing the input-to-hidden transformation matrix.
        B (nn.Parameter): The learnable parameter representing the hidden-to-output transformation matrix.

    Properties:
        weight (torch.Tensor): The weight matrix obtained by multiplying A and B and scaling it by the scale factor.

    Methods:
        forward(x): Applies the Lora transformation to the input tensor x.

    """

    def __init__(self, dim: int, dim_out: int, r: int = 8, alpha: float = 2):
        super().__init__()
        self.scale: float = alpha / r

        self.A: nn.Parameter = nn.Parameter(torch.randn(dim, r))
        self.B: nn.Parameter = nn.Parameter(torch.randn(r, dim_out))

    @property
    def weight(self) -> torch.Tensor:
        """Weight matrix obtained by multiplying A and B and scaling it by the scale factor.

        Returns:
            torch.Tensor: The weight matrix.
        """
        return (self.A @ self.B) * self.scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the Lora module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        return x @ self.weight



from typing import Tuple, Union

import torch.nn.functional as F
from einops import pack, unpack
from torch import nn

# helper


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def identity(t):
    return t


def divisible_by(num, den):
    return (num % den) == 0


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def is_odd(n):
    return not divisible_by(n, 2)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


# helper classes


class CausalConv3d(nn.Module):
    """
    Causal Convolution Module
    -------------------------

    This module is used to perform a causal convolution on a 3D tensor.
    It is used in the ResNet architecture to perform a causal convolution
    on the input tensor.

    Args:
        chan_in (int): The number of channels in the input tensor.
        chan_out (int): The number of channels in the output tensor.
        kernel_size (int): The size of the kernel used to perform the convolution.
        pad_mode (str): The padding mode used to pad the input tensor.
        kwargs (dict): Additional arguments to be passed to the convolution layer.

    Returns:
        out (tensor): The output tensor.

    Usage:
    >>> causal_conv = CausalConv3d(64, 128, 3)
    >>> x = torch.randn(1, 64, 32, 32)
    >>> out = causal_conv(x)



    """

    def __init__(
        self,
        chan_in,
        chan_out,
        kernel_size: Union[int, Tuple[int, int, int]],
        pad_mode="reflect",
        **kwargs,
    ):
        super().__init__()
        kernel_size = cast_tuple(kernel_size, 3)

        time_kernel_size, height_kernel_size, width_kernel_size = kernel_size

        assert is_odd(height_kernel_size) and is_odd(
            width_kernel_size
        ), "Height and width kernel sizes must be odd"

        dilation = kwargs.pop("dilation", 1)
        stride = kwargs.pop("stide", 2)

        self.pad_mode = pad_mode
        self.time_pad = dilation * (time_kernel_size - 1) + (1 - stride)
        height_pad = height_kernel_size // 2
        width_pad = width_kernel_size // 2

        self.time_causal_padding = (
            width_pad,
            width_pad,
            height_pad,
            height_pad,
            self.time_pad,
            0,
        )

        stride = (stride, 1, 1)
        dilation = (dilation, 1, 1)
        self.conv = nn.Conv3d(
            chan_in,
            chan_out,
            kernel_size,
            stride=stride,
            dilation=dilation,
            **kwargs,
        )

    def forward(self, x):
        """Forward pass of the module"""
        x = F.pad(x, self.time_causal_padding, mode=self.pad_mode)
        return self.conv(x)



import triton
import triton.language as tl


class Functions:
    @staticmethod
    @triton.jit
    def tanh_activation_kernel(
        x_ptr,
        out_ptr,
        n_elements: int,
        BLOCK_SIZE: tl.constexpr,
    ):
        """
        Applies the hyperbolic tangent (tanh) activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        exp2x = tl.exp(2 * x)
        output = 1 - 2 / (exp2x + 1)
        tl.store(out_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def hard_tanh_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the hard tanh activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        shape_condition = tl.where(x < -1, -1, x)
        output = tl.where(x > 1, 1, shape_condition)
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def relu_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the rectified linear unit (ReLU) activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = tl.maximum(0, x)
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def relu6_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the rectified linear unit 6 (ReLU 6) activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = tl.minimum(tl.maximum(x, 0), 6.0)
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def leaky_relu_activation_kernel(
        x_ptr, output_ptr, n_elements, alpha, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the LeakyReLU activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = tl.maximum(x, alpha * x)
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def smooth_relu_activation_kernel(
        x_ptr, output_ptr, n_elements, beta, BLOCK_SIZE: tl.constexpr
    ):
        """
        Convolution of ReLU with a box, transition region widens, the loss surface becomes smoother
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = tl.where(x >= beta, x, 0.0)
        output = tl.where(
            tl.abs(x) <= beta, ((x + beta) * (x + beta) / (4.0 * beta), output)
        )

        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def softsign_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the softsign activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = x / (tl.abs(x) + 1)
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def softplus_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the softplus activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = tl.log(1 + tl.exp(x))
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def sigmoid_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the sigmoid activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = 1 / (1 + tl.exp(-x))
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def hard_sigmoid_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the hard sigmoid activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        x_plus_3 = x + 3.0
        relu6_result = tl.minimum(tl.maximum(x_plus_3, 0), 6.0)
        output = relu6_result / 6.0
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def silu_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the Sigmoid-weighted Linear Unit (SiLU) activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        output = x * (1 / (1 + tl.exp(-x)))
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def hard_silu_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the hard SiLU activation function to element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        x_plus_3 = x + 3.0
        relu6_result = tl.minimum(tl.maximum(x_plus_3, 0), 6.0)
        hard_sigmoid_output = relu6_result / 6.0
        output = x * hard_sigmoid_output
        tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def softmax_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the softmax activation function to the input tensor along the specified axis
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        max_x = tl.maximum(x, 0)
        x -= max_x
        exp_x = tl.exp(x)
        sum_exp_x = tl.sum(exp_x)
        output = exp_x / sum_exp_x
        tl.store(output_ptr + offsets, output, mask=mask)

    @triton.jit
    def gelu_activation_kernel(
        x_ptr, output_ptr, approximation, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the Gaussian Error Linear Unit (GELU) activation function element-wise to the input tensor
        """
        idx = tl.program_id(0)
        block_st = idx * BLOCK_SIZE
        offsets = block_st + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        x = tl.load(x_ptr + offsets, mask=mask)

        if approximation is True:
            output = (
                0.5
                * x
                * (
                    1
                    + tl.libdevice.tanh(
                        tl.libdevice.sqrt(2.0 / 3.141592653589793)
                        * (x + 0.044715 * x * x * x)
                    )
                )
            )
            tl.store(output_ptr + offsets, output, mask=mask)
        else:
            output = x * 0.5 * (1.0 + tl.erf(x / tl.sqrt(2.0)))
            tl.store(output_ptr + offsets, output, mask=mask)

    @staticmethod
    @triton.jit
    def swiglu_activation_kernel(
        x_ptr, output_ptr, n_elements, BLOCK_SIZE: tl.constexpr
    ):
        """
        Applies the SwiGLU activation function to the input tensor
        """
        idx = tl.program_id(0) * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
        mask = idx < n_elements // 2
        f = tl.load(x_ptr + idx * 2, mask=mask)
        g = tl.load(x_ptr + idx * 2 + 1, mask=mask)
        g_silu = g * tl.sigmoid(g)
        output = f * g_silu

        tl.store(output_ptr + idx, output, mask=mask)



import math

import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from torch import einsum, nn
from torch.nn import LayerNorm

from zeta.utils.main import default, exists, l2norm


class CrossAttention(nn.Module):
    def __init__(
        self,
        dim,
        *,
        context_dim=None,
        dim_head=64,
        heads=8,
        dropout=0.0,
        norm_context=False,
        cosine_sim=False,
        cosine_sim_scale=16,
    ):
        """
        CrossAttention module performs cross-attention mechanism between input tensor `x` and context tensor `context`.

        Args:
            dim (int): The dimension of the input tensor `x`.
            context_dim (int, optional): The dimension of the context tensor `context`. If not provided, it defaults to `dim`.
            dim_head (int, optional): The dimension of each head in the multi-head attention. Defaults to 64.
            heads (int, optional): The number of attention heads. Defaults to 8.
            dropout (float, optional): The dropout rate. Defaults to 0.0.
            norm_context (bool, optional): Whether to apply layer normalization to the context tensor. Defaults to False.
            cosine_sim (bool, optional): Whether to use cosine similarity for attention calculation. Defaults to False.
            cosine_sim_scale (int, optional): The scale factor for cosine similarity. Defaults to 16.
        """
        super().__init__()
        self.cosine_sim = cosine_sim
        self.scale = cosine_sim_scale if cosine_sim else (dim_head**-0.5)
        self.heads = heads
        inner_dim = dim_head * heads

        context_dim = default(context_dim, dim)

        self.norm = LayerNorm(dim)
        self.norm_context = (
            LayerNorm(context_dim) if norm_context else nn.Identity()
        )
        self.dropout = nn.Dropout(dropout)

        self.null_kv = nn.Parameter(torch.randn(2, dim_head))
        self.to_q = nn.Linear(dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim, bias=False), LayerNorm(dim)
        )

    def forward(self, x, context, mask=None):
        """
        Forward pass of the CrossAttention module.

        Args:
            x (torch.Tensor): The input tensor of shape (batch_size, sequence_length, dim).
            context (torch.Tensor): The context tensor of shape (batch_size, context_length, context_dim).
            mask (torch.Tensor, optional): The attention mask tensor of shape (batch_size, sequence_length).

        Returns:
            torch.Tensor: The output tensor of shape (batch_size, sequence_length, dim).
        """
        b, _n, _device = *x.shape[:2], x.device

        x = self.norm(x)
        context = self.norm_context(context)

        q, k, v = (
            self.to_q(x),
            *self.to_kv(context).chunk(2, dim=-1),
        )

        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads),
            (q, k, v),
        )

        # add null key / value for classifier free guidance in prior net

        nk, nv = map(
            lambda t: repeat(t, "d -> b h 1 d", h=self.heads, b=b),
            self.null_kv.unbind(dim=-2),
        )

        k = torch.cat((nk, k), dim=-2)
        v = torch.cat((nv, v), dim=-2)

        if self.cosine_sim:
            q, k = map(l2norm, (q, k))

        q, k = map(lambda t: t * math.sqrt(self.scale), (q, k))

        sim = einsum("b h i d, b h j d -> b h i j", q, k)
        max_neg_value = -torch.finfo(sim.dtype).max

        if exists(mask):
            mask = F.pad(mask, (1, 0), value=True)
            mask = rearrange(mask, "b j -> b 1 1 j")
            sim = sim.masked_fill(~mask, max_neg_value)

        attn = sim.softmax(dim=-1, dtype=torch.float32)
        attn = attn.type(sim.dtype)

        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)



"""Transformer module."""

import math
from collections import namedtuple
from dataclasses import dataclass
from functools import partial, wraps
from inspect import isfunction
from random import random
from typing import Callable, List, Optional

import torch
import torch.nn.functional as F
from einops import rearrange, reduce, repeat
from packaging import version
from torch import Tensor, einsum, nn

# constants

EfficientAttentionConfig = namedtuple(
    "EfficientAttentionConfig",
    ["enable_flash", "enable_math", "enable_mem_efficient"],
)


@dataclass
class Intermediates:
    qk_similarities: Optional[Tensor] = None
    pre_softmax_attn: Optional[Tensor] = None
    post_softmax_attn: Optional[Tensor] = None

    def to_tuple(self):
        return (
            self.qk_similarities,
            self.pre_softmax_attn,
            self.post_softmax_attn,
        )


# helpers


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def compact(arr):
    return [*filter(exists, arr)]


def once(fn):
    called = False

    @wraps(fn)
    def inner(x):
        nonlocal called
        if called:
            return
        called = True
        return fn(x)

    return inner


print_once = once(print)

# functions for creating causal mask
# need a special one for onnx cpu (no support for .triu)


def create_causal_mask(i, j, device):
    return torch.ones((i, j), device=device, dtype=torch.bool).triu(j - i + 1)


def onnx_create_causal_mask(i, j, device):
    r = torch.arange(i, device=device)
    causal_mask = rearrange(r, "i -> i 1") < rearrange(r, "j -> 1 j")
    causal_mask = F.pad(causal_mask, (j - i, 0), value=False)
    return causal_mask


# main class


class Attend(nn.Module):
    """
    Attend module performs attention mechanism for neural networks.

    Args:
        dropout (float): Dropout probability. Default is 0.0.
        causal (bool): Whether to use causal attention. Default is False.
        heads (int): Number of attention heads. Default is None.
        talking_heads (bool): Whether to use talking heads attention. Default is False.
        sparse_topk (int): Number of top-k values to consider for sparse attention. Default is None.
        scale (float): Scaling factor for attention scores. Default is None.
        qk_norm (bool): Whether to normalize query-key dot products. Default is False.
        flash (bool): Whether to use flash attention. Default is False.
        add_zero_kv (bool): Whether to add a key/value token composed of zeros. Default is False.
        onnxable (bool): Whether the module is ONNX compatible. Default is False.
    """

    def __init__(
        self,
        *,
        dropout=0.0,
        causal=False,
        heads=None,
        talking_heads=False,
        sparse_topk=None,
        scale=None,
        qk_norm=False,
        flash=False,
        add_zero_kv=False,
        onnxable=False,
    ):
        super().__init__()
        self.scale = scale
        self.qk_norm = qk_norm

        self.causal = causal
        self.create_causal_mask = (
            onnx_create_causal_mask if onnxable else create_causal_mask
        )

        self.attn_fn = (
            partial(F.softmax, dtype=torch.float32)
            if not qk_norm
            else F.softmax
        )

        self.dropout = dropout
        self.attn_dropout = nn.Dropout(dropout)

        # talking heads

        assert not (
            flash and talking_heads
        ), "talking heads not compatible with flash attention"

        self.talking_heads = talking_heads
        if talking_heads:
            self.pre_softmax_talking_heads = nn.Conv2d(
                heads, heads, 1, bias=False
            )
            self.post_softmax_talking_heads = nn.Conv2d(
                heads, heads, 1, bias=False
            )

        # sparse topk

        assert not (
            flash and sparse_topk
        ), "sparse topk not compatible with flash attention"
        self.sparse_topk = sparse_topk

        # add a key / value token composed of zeros
        # in case this helps controlling outliers, proposed by
        # https://www.evanmiller.org/attention-is-off-by-one.html

        self.add_zero_kv = add_zero_kv

        # flash attention

        self.flash = flash
        assert not (
            flash and version.parse(torch.__version__) < version.parse("2.0.0")
        ), (
            "in order to use flash attention, you must be using pytorch 2.0 or"
            " above"
        )

        # determine efficient attention configs for cuda and cpu

        self.cpu_config = EfficientAttentionConfig(True, True, True)
        self.cuda_config = None

        if not torch.cuda.is_available() or not flash:
            return

        device_properties = torch.cuda.get_device_properties(
            torch.device("cuda")
        )

        if device_properties.major == 8 and device_properties.minor == 0:
            print_once(
                "A100 GPU detected, using flash attention if input tensor is on"
                " cuda"
            )
            self.cuda_config = EfficientAttentionConfig(True, False, False)
        else:
            print_once(
                "Non-A100 GPU detected, using math or mem efficient attention"
                " if input tensor is on cuda"
            )
            self.cuda_config = EfficientAttentionConfig(False, True, True)

    def flash_attn(self, q, k, v, mask=None, attn_bias=None):
        """
        Perform flash attention.

        Args:
            q (torch.Tensor): Query tensor.
            k (torch.Tensor): Key tensor.
            v (torch.Tensor): Value tensor.
            mask (torch.Tensor): Mask tensor. Default is None.
            attn_bias (torch.Tensor): Attention bias tensor. Default is None.

        Returns:
            torch.Tensor: Output tensor.
            Intermediates: Intermediate values during attention computation.
        """

        batch, heads, q_len, _, k_len, is_cuda, device = (
            *q.shape,
            k.shape[-2],
            q.is_cuda,
            q.device,
        )

        # Recommended for multi-query single-key-value attention by Tri Dao
        # kv shape torch.Size([1, 512, 64]) -> torch.Size([1, 8, 512, 64])

        if k.ndim == 3:
            k = rearrange(k, "b ... -> b 1 ...").expand_as(q)

        if v.ndim == 3:
            v = rearrange(v, "b ... -> b 1 ...").expand_as(q)

        # handle scale - by default they scale by dim_head ** -0.5, but need to
        # take care if using cosine sim attention

        if self.qk_norm:
            default_scale = q.shape[-1] ** -0.5
            q = q * (default_scale / self.scale)

        # Check if mask exists and expand to compatible shape
        # The mask is B L, so it would have to be expanded to B H N L

        causal = self.causal

        if exists(mask):
            assert mask.ndim == 4
            mask = mask.expand(batch, heads, q_len, k_len)

            # manually handle causal mask, if another mask was given

            if causal:
                causal_mask = self.create_causal_mask(
                    q_len, k_len, device=device
                )
                mask = mask & ~causal_mask
                causal = False

        # handle alibi positional bias
        # convert from bool to float

        if exists(attn_bias):
            attn_bias = rearrange(attn_bias, "h i j -> 1 h i j").expand(
                batch, heads, -1, -1
            )

            # if mask given, the mask would already contain the causal mask from above logic
            # otherwise, if no mask given but still causal, mask out alibi
            # positional bias to a large negative number

            mask_value = -torch.finfo(q.dtype).max

            if exists(mask):
                attn_bias = attn_bias.masked_fill(~mask, mask_value // 2)
            elif causal:
                causal_mask = self.create_causal_mask(
                    q_len, k_len, device=device
                )
                attn_bias = attn_bias.masked_fill(causal_mask, mask_value // 2)
                causal = False

            # scaled_dot_product_attention handles attn_mask either as bool or additive bias
            # make it an additive bias here

            mask = attn_bias

        # Check if there is a compatible device for flash attention

        config = self.cuda_config if is_cuda else self.cpu_config

        # pytorch 2.0 flash attn: q, k, v, mask, dropout, causal, softmax_scale

        with torch.backends.cuda.sdp_kernel(**config._asdict()):
            out = F.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=mask,
                dropout_p=self.dropout if self.training else 0.0,
                is_causal=causal,
            )

        return out, Intermediates()

    def forward(self, q, k, v, mask=None, attn_bias=None, prev_attn=None):
        """
        Perform forward pass of the Attend module.

        Args:
            q (torch.Tensor): Query tensor.
            k (torch.Tensor): Key tensor.
            v (torch.Tensor): Value tensor.
            mask (torch.Tensor): Mask tensor. Default is None.
            attn_bias (torch.Tensor): Attention bias tensor. Default is None.
            prev_attn (torch.Tensor): Previous attention tensor. Default is None.

        Returns:
            torch.Tensor: Output tensor.
            Intermediates: Intermediate values during attention computation.
        """

        _n, heads, kv_heads, device = (
            q.shape[-2],
            q.shape[1],
            k.shape[1],
            q.device,
        )

        scale = default(self.scale, q.shape[-1] ** -0.5)

        # handle grouped multi-query attention

        if kv_heads == 1:
            k, v = map(lambda t: rearrange(t, "b 1 n d -> b n d"), (k, v))
        elif kv_heads < heads:
            k, v = map(
                lambda t: repeat(
                    t, "b kvh n d -> b (r kvh) n d", r=heads // kv_heads
                ),
                (k, v),
            )

        # handle zero kv, as means for allowing network to attend to nothing

        if self.add_zero_kv:
            k, v = map(lambda t: F.pad(t, (0, 0, 1, 0), value=0.0), (k, v))

            if exists(mask):
                mask = F.pad(mask, (1, 0), value=True)

            if exists(attn_bias):
                attn_bias = F.pad(attn_bias, (1, 0), value=0.0)

        if self.flash:
            assert not exists(
                prev_attn
            ), "residual attention not compatible with flash attention"
            return self.flash_attn(q, k, v, mask=mask, attn_bias=attn_bias)

        kv_einsum_eq = "b j d" if k.ndim == 3 else "b h j d"

        dots = einsum(f"b h i d, {kv_einsum_eq} -> b h i j", q, k) * scale

        if exists(prev_attn):
            dots = dots + prev_attn

        qk_similarities = dots.clone()

        if self.talking_heads:
            dots = self.pre_softmax_talking_heads(dots)

        if exists(attn_bias):
            dots = dots + attn_bias

        i, j, dtype = *dots.shape[-2:], dots.dtype

        mask_value = -torch.finfo(dots.dtype).max

        if exists(self.sparse_topk) and self.sparse_topk < j:
            top_values, _ = dots.topk(self.sparse_topk, dim=-1)
            sparse_topk_mask = dots < top_values[..., -1:]
            mask = (
                (mask & sparse_topk_mask) if exists(mask) else sparse_topk_mask
            )

        if exists(mask):
            dots = dots.masked_fill(~mask, mask_value)

        if self.causal:
            causal_mask = self.create_causal_mask(i, j, device=device)
            dots = dots.masked_fill(causal_mask, mask_value)

        pre_softmax_attn = dots.clone()

        attn = self.attn_fn(dots, dim=-1)
        attn = attn.type(dtype)

        post_softmax_attn = attn.clone()

        attn = self.attn_dropout(attn)

        if self.talking_heads:
            attn = self.post_softmax_talking_heads(attn)

        out = einsum(f"b h i j, {kv_einsum_eq} -> b h i d", attn, v)

        intermediates = Intermediates(
            qk_similarities=qk_similarities,
            pre_softmax_attn=pre_softmax_attn,
            post_softmax_attn=post_softmax_attn,
        )

        return out, intermediates


# cascading heads logic


def to_single_heads(t, dim=1):
    heads = t.unbind(dim=dim)
    return tuple(head.unsqueeze(dim) for head in heads)


class CascadingHeads(nn.Module):
    def __init__(self, attend: Attend):
        super().__init__()
        self.attend = attend

    def forward(self, q, k, v, mask=None, attn_bias=None, prev_attn=None):
        assert q.shape[-1] == v.shape[-1], (
            "cascading heads can only be done if query / key and value head"
            " dimensions are the same"
        )

        # split inputs into per-head inputs

        heads = q.shape[1]

        queries = to_single_heads(q)
        keys = to_single_heads(k) if k.ndim == 4 else ((k,) * heads)
        values = to_single_heads(v) if v.ndim == 4 else ((v,) * heads)

        mask = (mask,) * heads

        attn_bias = (
            to_single_heads(attn_bias, dim=0)
            if exists(attn_bias)
            else ((None,) * heads)
        )
        prev_attn = (
            to_single_heads(prev_attn)
            if exists(prev_attn)
            else ((None,) * heads)
        )

        # now loop through each head, without output of previous head summed with the next head
        # thus cascading

        all_outs = []
        all_intermediates = []

        prev_head_out = None

        for h_q, h_k, h_v, h_mask, h_attn_bias, h_prev_attn in zip(
            queries, keys, values, mask, attn_bias, prev_attn
        ):
            if exists(prev_head_out):
                h_q = h_q + prev_head_out

            out, intermediates = self.attend(
                h_q,
                h_k,
                h_v,
                mask=h_mask,
                attn_bias=h_attn_bias,
                prev_attn=h_prev_attn,
            )

            prev_head_out = out

            all_outs.append(out)
            all_intermediates.append(intermediates)

        # cat all output heads

        all_outs = torch.cat(all_outs, dim=1)

        # cat all intermediates, if they exist

        qk_similarities, pre_softmax_attn, post_softmax_attn = zip(
            *map(lambda i: i.to_tuple(), all_intermediates)
        )

        qk_similarities, pre_softmax_attn, post_softmax_attn = map(
            compact, (qk_similarities, pre_softmax_attn, post_softmax_attn)
        )

        aggregated_intermediates = Intermediates(
            qk_similarities=(
                torch.cat(qk_similarities, dim=1)
                if len(qk_similarities) > 0
                else None
            ),
            pre_softmax_attn=(
                torch.cat(pre_softmax_attn, dim=1)
                if len(pre_softmax_attn) > 0
                else None
            ),
            post_softmax_attn=(
                torch.cat(post_softmax_attn, dim=1)
                if len(post_softmax_attn) > 0
                else None
            ),
        )

        return all_outs, aggregated_intermediates


# Utils
EfficientAttentionConfig = namedtuple(
    "EfficientAttentionConfig",
    ["enable_flash", "enable_math", "enable_mem_efficient"],
)

DEFAULT_DIM_HEAD = 64


@dataclass
class LayerIntermediates:
    hiddens: Optional[List[Tensor]] = None
    attn_intermediates: Optional[List[Intermediates]] = None
    layer_hiddens: Optional[List[Tensor]] = None
    attn_z_loss: Optional[Tensor] = None


# helpers


def exists(val):
    return val is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


def cast_tuple(val, depth):
    return val if isinstance(val, tuple) else (val,) * depth


def divisible_by(num, den):
    return (num % den) == 0


def maybe(fn):
    @wraps(fn)
    def inner(x, *args, **kwargs):
        if not exists(x):
            return x
        return fn(x, *args, **kwargs)

    return inner


class always:
    def __init__(self, val):
        self.val = val

    def __call__(self, *args, **kwargs):
        return self.val


class not_equals:
    def __init__(self, val):
        self.val = val

    def __call__(self, x, *args, **kwargs):
        return x != self.val


class equals:
    def __init__(self, val):
        self.val = val

    def __call__(self, x, *args, **kwargs):
        return x == self.val


def Sequential(*modules):
    return nn.Sequential(*filter(exists, modules))


# tensor helpers


def max_neg_value(tensor):
    return -torch.finfo(tensor.dtype).max


def l2norm(t, groups=1):
    t = rearrange(t, "... (g d) -> ... g d", g=groups)
    t = F.normalize(t, p=2, dim=-1)
    return rearrange(t, "... g d -> ... (g d)")


def pad_at_dim(t, pad, dim=-1, value=0.0):
    dims_from_right = (-dim - 1) if dim < 0 else (t.ndim - dim - 1)
    zeros = (0, 0) * dims_from_right
    return F.pad(t, (*zeros, *pad), value=value)


def or_reduce(masks):
    head, *body = masks
    for rest in body:
        head = head | rest
    return head


# auxiliary loss helpers


def calc_z_loss(pre_softmax_attns: List[Tensor], mask=None, weight=1.0):
    # the same loss applied to the mixture of experts router logits in https://arxiv.org/abs/2202.08906
    # in the paper, in a tiny footnote, they mention using it on attention logits with stabilizing effects
    # also used in PaLM as one of the measures

    lse = 0.0

    for attn in pre_softmax_attns:
        lse = lse + attn.logsumexp(dim=-1)

    loss = torch.square(lse)
    loss = reduce(loss, "b h n -> b n", "sum")

    if not exists(mask):
        return loss.mean() * weight

    loss = loss[mask].sum() / mask.sum().clamp(min=1e-5)
    return loss * weight


# init helpers


def init_zero_(layer):
    nn.init.constant_(layer.weight, 0.0)
    if exists(layer.bias):
        nn.init.constant_(layer.bias, 0.0)


# keyword argument helpers


def pick_and_pop(keys, d):
    values = list(map(d.pop, keys))
    return dict(zip(keys, values))


def group_dict_by_key(cond, d):
    return_val = [{}, {}]
    for key in d.keys():
        match = bool(cond(key))
        ind = int(not match)
        return_val[ind][key] = d[key]
    return (*return_val,)


def string_begins_with(prefix, str):
    return str.startswith(prefix)


def group_by_key_prefix(prefix, d):
    return group_dict_by_key(partial(string_begins_with, prefix), d)


def groupby_prefix_and_trim(prefix, d):
    kwargs_with_prefix, kwargs = group_dict_by_key(
        partial(string_begins_with, prefix), d
    )
    kwargs_without_prefix = dict(
        map(
            lambda x: (x[0][len(prefix) :], x[1]),
            tuple(kwargs_with_prefix.items()),
        )
    )
    return kwargs_without_prefix, kwargs


# initializations


def deepnorm_init(
    transformer, beta, module_name_match_list=[".ff.", ".to_v", ".to_out"]
):
    for name, module in transformer.named_modules():
        if not isinstance(module, nn.Linear):
            continue

        needs_beta_gain = any(
            map(lambda substr: substr in name, module_name_match_list)
        )
        gain = beta if needs_beta_gain else 1
        nn.init.xavier_normal_(module.weight.data, gain=gain)

        if exists(module.bias):
            nn.init.constant_(module.bias.data, 0)


# structured dropout, more effective than traditional attention dropouts


def dropout_seq(seq, mask, dropout):
    b, n, *_, device = *seq.shape, seq.device
    logits = torch.randn(b, n, device=device)

    if exists(mask):
        mask_value = max_neg_value(logits)
        logits = logits.masked_fill(~mask, mask_value)

    keep_prob = 1.0 - dropout
    num_keep = max(1, int(keep_prob * n))
    keep_indices = logits.topk(num_keep, dim=1).indices

    batch_indices = torch.arange(b, device=device)
    batch_indices = rearrange(batch_indices, "b -> b 1")

    seq = seq[batch_indices, keep_indices]

    if exists(mask):
        seq_counts = mask.sum(dim=-1)
        seq_keep_counts = torch.ceil(seq_counts * keep_prob).int()
        keep_mask = torch.arange(num_keep, device=device) < rearrange(
            seq_keep_counts, "b -> b 1"
        )

        mask = mask[batch_indices, keep_indices] & keep_mask

    return seq, mask


# activations


class ReluSquared(nn.Module):
    def forward(self, x):
        return F.relu(x) ** 2


# embedding


class TokenEmbedding(nn.Module):
    def __init__(self, dim, num_tokens, l2norm_embed=False):
        super().__init__()
        self.l2norm_embed = l2norm_embed
        self.emb = nn.Embedding(num_tokens, dim)

    def forward(self, x):
        token_emb = self.emb(x)
        return l2norm(token_emb) if self.l2norm_embed else token_emb


# positional embeddings


class AbsolutePositionalEmbedding(nn.Module):
    def __init__(self, dim, max_seq_len, l2norm_embed=False):
        super().__init__()
        self.scale = dim**-0.5 if not l2norm_embed else 1.0
        self.max_seq_len = max_seq_len
        self.l2norm_embed = l2norm_embed
        self.emb = nn.Embedding(max_seq_len, dim)

    def forward(self, x, pos=None):
        seq_len, device = x.shape[1], x.device
        assert seq_len <= self.max_seq_len, (
            f"you are passing in a sequence length of {seq_len} but your"
            " absolute positional embedding has a max sequence length of"
            f" {self.max_seq_len}"
        )

        if not exists(pos):
            pos = torch.arange(seq_len, device=device)

        pos_emb = self.emb(pos)
        pos_emb = pos_emb * self.scale
        return l2norm(pos_emb) if self.l2norm_embed else pos_emb


class ScaledSinusoidalEmbedding(nn.Module):
    def __init__(self, dim, theta=10000):
        super().__init__()
        assert divisible_by(dim, 2)
        self.scale = nn.Parameter(torch.ones(1) * dim**-0.5)

        half_dim = dim // 2
        freq_seq = torch.arange(half_dim).float() / half_dim
        inv_freq = theta**-freq_seq
        self.register_buffer("inv_freq", inv_freq, persistent=False)

    def forward(self, x, pos=None):
        seq_len, device = x.shape[1], x.device

        if not exists(pos):
            pos = torch.arange(seq_len, device=device)

        emb = einsum("i, j -> i j", pos, self.inv_freq)
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb * self.scale


class RelativePositionBias(nn.Module):
    def __init__(
        self, scale, causal=False, num_buckets=32, max_distance=128, heads=8
    ):
        super().__init__()
        self.scale = scale
        self.causal = causal
        self.num_buckets = num_buckets
        self.max_distance = max_distance
        self.relative_attention_bias = nn.Embedding(num_buckets, heads)

    @staticmethod
    def _relative_position_bucket(
        relative_position, causal=True, num_buckets=32, max_distance=128
    ):
        ret = 0
        n = -relative_position
        if not causal:
            num_buckets //= 2
            ret += (n < 0).long() * num_buckets
            n = torch.abs(n)
        else:
            n = torch.max(n, torch.zeros_like(n))

        max_exact = num_buckets // 2
        is_small = n < max_exact

        val_if_large = (
            max_exact
            + (
                torch.log(n.float() / max_exact)
                / math.log(max_distance / max_exact)
                * (num_buckets - max_exact)
            ).long()
        )
        val_if_large = torch.min(
            val_if_large, torch.full_like(val_if_large, num_buckets - 1)
        )

        ret += torch.where(is_small, n, val_if_large)
        return ret

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, i, j):
        device = self.device
        q_pos = torch.arange(j - i, j, dtype=torch.long, device=device)
        k_pos = torch.arange(j, dtype=torch.long, device=device)
        rel_pos = k_pos[None, :] - q_pos[:, None]
        rp_bucket = self._relative_position_bucket(
            rel_pos,
            causal=self.causal,
            num_buckets=self.num_buckets,
            max_distance=self.max_distance,
        )
        values = self.relative_attention_bias(rp_bucket)
        bias = rearrange(values, "i j h -> h i j")
        return bias * self.scale


class DynamicPositionBias(nn.Module):
    def __init__(self, dim, *, heads, depth, log_distance=False, norm=False):
        super().__init__()
        assert (
            depth >= 1
        ), "depth for dynamic position bias MLP must be greater or equal to 1"
        self.log_distance = log_distance

        self.mlp = nn.ModuleList([])

        self.mlp.append(
            Sequential(
                nn.Linear(1, dim),
                nn.LayerNorm(dim) if norm else None,
                nn.SiLU(),
            )
        )

        for _ in range(depth - 1):
            self.mlp.append(
                Sequential(
                    nn.Linear(dim, dim),
                    nn.LayerNorm(dim) if norm else None,
                    nn.SiLU(),
                )
            )

        self.mlp.append(nn.Linear(dim, heads))

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, i, j):
        assert i == j
        n, device = j, self.device

        # get the (n x n) matrix of distances
        seq_arange = torch.arange(n, device=device)
        context_arange = torch.arange(n, device=device)
        indices = rearrange(seq_arange, "i -> i 1") - rearrange(
            context_arange, "j -> 1 j"
        )
        indices += n - 1

        # input to continuous positions MLP
        pos = torch.arange(-n + 1, n, device=device).float()
        pos = rearrange(pos, "... -> ... 1")

        if self.log_distance:
            # log of distance is sign(rel_pos) * log(abs(rel_pos) + 1)
            pos = torch.sign(pos) * torch.log(pos.abs() + 1)

        for layer in self.mlp:
            pos = layer(pos)

        # get position biases
        bias = pos[indices]
        bias = rearrange(bias, "i j h -> h i j")
        return bias


class AlibiPositionalBias(nn.Module):
    def __init__(self, heads, total_heads, **kwargs):
        super().__init__()
        self.heads = heads
        self.total_heads = total_heads

        slopes = Tensor(self._get_slopes(heads))
        slopes = rearrange(slopes, "h -> h 1 1")
        self.register_buffer("slopes", slopes, persistent=False)
        self.register_buffer("bias", None, persistent=False)

    def get_bias(self, i, j, device):
        i_arange = torch.arange(j - i, j, device=device)
        j_arange = torch.arange(j, device=device)
        bias = -torch.abs(
            rearrange(j_arange, "j -> 1 1 j")
            - rearrange(i_arange, "i -> 1 i 1")
        )
        return bias

    @staticmethod
    def _get_slopes(heads):
        def get_slopes_power_of_2(n):
            start = 2 ** (-(2 ** -(math.log2(n) - 3)))
            ratio = start
            return [start * ratio**i for i in range(n)]

        if math.log2(heads).is_integer():
            return get_slopes_power_of_2(heads)

        closest_power_of_2 = 2 ** math.floor(math.log2(heads))
        return (
            get_slopes_power_of_2(closest_power_of_2)
            + get_slopes_power_of_2(2 * closest_power_of_2)[0::2][
                : heads - closest_power_of_2
            ]
        )

    @property
    def device(self):
        return next(self.buffers()).device

    def forward(self, i, j):
        h, device = self.total_heads, self.device

        if (
            exists(self.bias)
            and self.bias.shape[-1] >= j
            and self.bias.shape[-2] >= i
        ):
            return self.bias[..., :i, :j]

        bias = self.get_bias(i, j, device)
        bias = bias * self.slopes

        num_heads_unalibied = h - bias.shape[0]
        bias = pad_at_dim(bias, (0, num_heads_unalibied), dim=0)
        self.register_buffer("bias", bias, persistent=False)

        return self.bias


class RotaryEmbedding(nn.Module):
    def __init__(
        self,
        dim,
        use_xpos=False,
        scale_base=512,
        interpolation_factor=1.0,
        base=10000,
        base_rescale_factor=1.0,
    ):
        super().__init__()
        # proposed by reddit user bloc97, to rescale rotary embeddings to longer sequence length without fine-tuning
        # has some connection to NTK literature
        # https://www.reddit.com/r/LocalLLaMA/comments/14lz7j5/ntkaware_scaled_rope_allows_llama_models_to_have/
        base *= base_rescale_factor ** (dim / (dim - 2))

        inv_freq = 1.0 / (base ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

        assert interpolation_factor >= 1.0
        self.interpolation_factor = interpolation_factor

        if not use_xpos:
            self.register_buffer("scale", None)
            return

        scale = (torch.arange(0, dim, 2) + 0.4 * dim) / (1.4 * dim)

        self.scale_base = scale_base
        self.register_buffer("scale", scale)

    def forward(self, seq_len, device):
        t = torch.arange(seq_len, device=device).type_as(self.inv_freq)
        t = t / self.interpolation_factor

        freqs = torch.einsum("i , j -> i j", t, self.inv_freq)
        freqs = torch.cat((freqs, freqs), dim=-1)

        if not exists(self.scale):
            return freqs, 1.0

        power = (
            torch.arange(seq_len, device=device) - (seq_len // 2)
        ) / self.scale_base
        scale = self.scale ** rearrange(power, "n -> n 1")
        scale = torch.cat((scale, scale), dim=-1)

        return freqs, scale


def rotate_half(x):
    x = rearrange(x, "... (j d) -> ... j d", j=2)
    x1, x2 = x.unbind(dim=-2)
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_pos_emb(t, freqs, scale=1):
    seq_len = t.shape[-2]
    freqs = freqs[-seq_len:, :]
    return (t * freqs.cos() * scale) + (rotate_half(t) * freqs.sin() * scale)


# norms


class Scale(nn.Module):
    def __init__(self, value, fn):
        super().__init__()
        self.value = value
        self.fn = fn

    def forward(self, x, **kwargs):
        out = self.fn(x, **kwargs)

        def scale_fn(t):
            return t * self.value

        if not isinstance(out, tuple):
            return scale_fn(out)

        return (scale_fn(out[0]), *out[1:])


class ScaleNorm(nn.Module):
    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.g = nn.Parameter(torch.ones(1) * (dim**-0.5))

    def forward(self, x):
        norm = torch.norm(x, dim=-1, keepdim=True)
        return x / norm.clamp(min=self.eps) * self.g


class RMSNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.scale = dim**0.5
        self.g = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        return F.normalize(x, dim=-1) * self.scale * self.g


class SimpleRMSNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.scale = dim**0.5

    def forward(self, x):
        return F.normalize(x, dim=-1) * self.scale


# residual and residual gates


class Residual(nn.Module):
    def __init__(self, dim, scale_residual=False, scale_residual_constant=1.0):
        super().__init__()
        self.residual_scale = (
            nn.Parameter(torch.ones(dim)) if scale_residual else None
        )
        self.scale_residual_constant = scale_residual_constant

    def forward(self, x, residual):
        if exists(self.residual_scale):
            residual = residual * self.residual_scale

        if self.scale_residual_constant != 1:
            residual = residual * self.scale_residual_constant

        return x + residual


class GRUGating(nn.Module):
    def __init__(self, dim, scale_residual=False, **kwargs):
        super().__init__()
        self.gru = nn.GRUCell(dim, dim)
        self.residual_scale = (
            nn.Parameter(torch.ones(dim)) if scale_residual else None
        )

    def forward(self, x, residual):
        if exists(self.residual_scale):
            residual = residual * self.residual_scale

        gated_output = self.gru(
            rearrange(x, "b n d -> (b n) d"),
            rearrange(residual, "b n d -> (b n) d"),
        )

        return gated_output.reshape_as(x)


# token shifting


def shift(t, amount, mask=None):
    if amount == 0:
        return t
    else:
        amount = min(amount, t.shape[1])

    if exists(mask):
        t = t.masked_fill(~mask[..., None], 0.0)

    return pad_at_dim(t, (amount, -amount), dim=-2, value=0.0)


class ShiftTokens(nn.Module):
    def __init__(self, shifts, fn):
        super().__init__()
        self.fn = fn
        self.shifts = tuple(shifts)

    def forward(self, x, **kwargs):
        mask = kwargs.get("mask", None)
        shifts = self.shifts
        segments = len(shifts)
        feats_per_shift = x.shape[-1] // segments
        splitted = x.split(feats_per_shift, dim=-1)
        segments_to_shift, rest = splitted[:segments], splitted[segments:]
        segments_to_shift = list(
            map(
                lambda args: shift(*args, mask=mask),
                zip(segments_to_shift, shifts),
            )
        )
        x = torch.cat((*segments_to_shift, *rest), dim=-1)
        return self.fn(x, **kwargs)


# feedforward


class GLU(nn.Module):
    def __init__(self, dim_in, dim_out, activation: Callable, mult_bias=False):
        super().__init__()
        self.act = activation
        self.proj = nn.Linear(dim_in, dim_out * 2)
        self.mult_bias = nn.Parameter(torch.ones(dim_out)) if mult_bias else 1.0

    def forward(self, x):
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * self.act(gate) * self.mult_bias


class FeedForward(nn.Module):
    def __init__(
        self,
        dim,
        dim_out=None,
        mult=4,
        glu=False,
        glu_mult_bias=False,
        swish=False,
        relu_squared=False,
        post_act_ln=False,
        dropout=0.0,
        no_bias=False,
        zero_init_output=False,
    ):
        super().__init__()
        inner_dim = int(dim * mult)
        dim_out = default(dim_out, dim)

        if relu_squared:
            activation = ReluSquared()
        elif swish:
            activation = nn.SiLU()
        else:
            activation = nn.GELU()

        if glu:
            project_in = GLU(
                dim, inner_dim, activation, mult_bias=glu_mult_bias
            )
        else:
            project_in = nn.Sequential(
                nn.Linear(dim, inner_dim, bias=not no_bias), activation
            )

        self.ff = Sequential(
            project_in,
            nn.LayerNorm(inner_dim) if post_act_ln else None,
            nn.Dropout(dropout),
            nn.Linear(inner_dim, dim_out, bias=not no_bias),
        )

        # init last linear layer to 0
        if zero_init_output:
            init_zero_(self.ff[-1])

    def forward(self, x):
        return self.ff(x)


# attention. it is all we need


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        dim_head=DEFAULT_DIM_HEAD,
        heads=8,
        causal=False,
        flash=False,
        talking_heads=False,
        head_scale=False,
        sparse_topk=None,
        num_mem_kv=0,
        dropout=0.0,
        on_attn=False,
        gate_values=False,
        zero_init_output=False,
        max_attend_past=None,
        qk_norm=False,
        qk_norm_groups=1,
        qk_norm_scale=10,
        qk_norm_dim_scale=False,
        one_kv_head=False,
        kv_heads=None,
        shared_kv=False,
        value_dim_head=None,
        tensor_product=False,  # https://arxiv.org/abs/2208.06061
        cascading_heads=False,
        add_zero_kv=False,  # same as add_zero_attn in pytorch
        onnxable=False,
    ):
        super().__init__()
        self.scale = dim_head**-0.5

        self.heads = heads
        self.causal = causal
        self.max_attend_past = max_attend_past

        assert not (exists(kv_heads) and one_kv_head), (
            "either attn_one_kv_head is set to True (in which case kv_heads is"
            " set to 1), or attn_kv_heads is set, but not both"
        )

        value_dim_head = default(value_dim_head, dim_head)
        kv_heads = default(kv_heads, heads)

        kv_heads = 1 if one_kv_head else kv_heads
        assert divisible_by(heads, kv_heads)

        self.kv_heads = kv_heads

        q_dim = dim_head * heads
        k_dim = dim_head * kv_heads
        v_dim = value_dim_head * kv_heads
        out_dim = value_dim_head * heads

        self.to_q = nn.Linear(dim, q_dim, bias=False)
        self.to_k = nn.Linear(dim, k_dim, bias=False)

        # shared key / values, for further memory savings during inference
        assert not (
            shared_kv and value_dim_head != dim_head
        ), "key and value head dimensions must be equal for shared key / values"
        self.to_v = nn.Linear(dim, v_dim, bias=False) if not shared_kv else None

        # relations projection from tp-attention
        self.to_r = (
            nn.Linear(dim, v_dim, bias=False) if tensor_product else None
        )

        # add GLU gating for aggregated values, from alphafold2
        self.to_v_gate = None
        if gate_values:
            self.to_v_gate = nn.Linear(dim, out_dim)
            nn.init.constant_(self.to_v_gate.weight, 0)
            nn.init.constant_(self.to_v_gate.bias, 1)

        # cosine sim attention
        self.qk_norm = qk_norm
        self.qk_norm_groups = qk_norm_groups
        self.qk_norm_scale = qk_norm_scale

        # whether to use the rmsnorm (equivalent to cosine sim attention when
        # scale is equal to 1) - https://arxiv.org/abs/2302.05442
        self.qk_norm_dim_scale = qk_norm_dim_scale

        self.qk_norm_q_scale = self.qk_norm_k_scale = 1
        if qk_norm and qk_norm_dim_scale:
            self.qk_norm_q_scale = nn.Parameter(torch.ones(dim_head))
            self.qk_norm_k_scale = nn.Parameter(torch.ones(dim_head))

        assert (not qk_norm) or divisible_by(dim_head, qk_norm_groups), (
            "dimension per attention head must be divisible by the qk norm"
            " groups"
        )
        assert not (qk_norm and (dim_head // qk_norm_groups) <= 2), (
            "the group dimension may be too small (2 was too small in my tests,"
            " but 4 still works, surprisingly)"
        )

        # attend class - includes core attention algorithm + talking heads

        self.attend = Attend(
            heads=heads,
            causal=causal,
            talking_heads=talking_heads,
            dropout=dropout,
            sparse_topk=sparse_topk,
            qk_norm=qk_norm,
            scale=qk_norm_scale if qk_norm else self.scale,
            add_zero_kv=add_zero_kv,
            flash=flash,
            onnxable=onnxable,
        )

        # if cascading_heads:
        #     # cascading heads - wrap the Attend logic
        #     self.attend = CascadingHeads(self.attend)

        # head scaling
        self.head_scale = head_scale
        if head_scale:
            self.head_scale_params = nn.Parameter(torch.ones(1, heads, 1, 1))

        # explicit topk sparse attention
        self.sparse_topk = sparse_topk

        # add memory key / values
        self.num_mem_kv = num_mem_kv
        if num_mem_kv > 0:
            self.mem_k = nn.Parameter(torch.randn(heads, num_mem_kv, dim_head))
            self.mem_v = nn.Parameter(torch.randn(heads, num_mem_kv, dim_head))

        # attention on attention
        self.attn_on_attn = on_attn
        self.to_out = (
            nn.Sequential(nn.Linear(out_dim, dim * 2, bias=False), nn.GLU())
            if on_attn
            else nn.Linear(out_dim, dim, bias=False)
        )

        # init output projection 0
        if zero_init_output:
            init_zero_(self.to_out)

    def forward(
        self,
        x,
        context=None,
        mask=None,
        context_mask=None,
        attn_mask=None,
        rel_pos=None,
        rotary_pos_emb=None,
        prev_attn=None,
        mem=None,
    ):
        # einops conflicts with ruff, so noqa on next line
        b, n, _, h, kv_h, head_scale, device, has_context = (  # noqa F841
            *x.shape,
            self.heads,
            self.kv_heads,
            self.head_scale,
            x.device,
            exists(context),
        )
        kv_input = default(context, x)

        q_input = x
        k_input = kv_input
        v_input = kv_input
        r_input = x

        if exists(mem):
            k_input = torch.cat((mem, k_input), dim=-2)
            v_input = torch.cat((mem, v_input), dim=-2)

        q = self.to_q(q_input)
        k = self.to_k(k_input)
        v = self.to_v(v_input) if exists(self.to_v) else k
        r = self.to_r(r_input) if exists(self.to_r) else None

        q = rearrange(q, "b n (h d) -> b h n d", h=h)

        k, v, r = map(
            lambda t: maybe(rearrange)(t, "b n (h d) -> b h n d", h=kv_h),
            (k, v, r),
        )

        if self.qk_norm:
            qk_l2norm = partial(l2norm, groups=self.qk_norm_groups)
            q, k = map(qk_l2norm, (q, k))

            q = q * self.qk_norm_q_scale
            k = k * self.qk_norm_k_scale

        if exists(rotary_pos_emb) and not has_context:
            freqs, xpos_scale = rotary_pos_emb
            l = freqs.shape[-1]  # noqa F741

            q_xpos_scale, k_xpos_scale = (
                (xpos_scale, xpos_scale**-1.0)
                if exists(xpos_scale)
                else (1.0, 1.0)
            )
            (ql, qr), (kl, kr), (vl, vr) = map(
                lambda t: (t[..., :l], t[..., l:]), (q, k, v)
            )

            ql, kl, vl = map(
                lambda arg: apply_rotary_pos_emb(arg[0], freqs, arg[1]),
                ((ql, q_xpos_scale), (kl, k_xpos_scale), (vl, k_xpos_scale)),
            )
            q, k, v = map(
                lambda t: torch.cat(t, dim=-1), ((ql, qr), (kl, kr), (vl, vr))
            )

        input_mask = context_mask if has_context else mask

        if self.num_mem_kv > 0:
            mem_k, mem_v = map(
                lambda t: repeat(t, "h n d -> b h n d", b=b),
                (self.mem_k, self.mem_v),
            )

            if self.qk_norm:
                mem_k = l2norm(mem_k)
                mem_k = mem_k * self.qk_norm_k_scale

            k = torch.cat((mem_k, k), dim=-2)
            v = torch.cat((mem_v, v), dim=-2)

            if exists(input_mask):
                input_mask = pad_at_dim(
                    input_mask, (self.num_mem_kv, 0), dim=-1, value=True
                )

        i, j = map(lambda t: t.shape[-2], (q, k))

        # determine masking

        max_neg_value(q)
        masks = []
        final_attn_mask = None

        if exists(input_mask):
            input_mask = rearrange(input_mask, "b j -> b 1 1 j")
            masks.append(~input_mask)

        if exists(attn_mask):
            assert 2 <= attn_mask.ndim <= 4, (
                "attention mask must have greater than 2 dimensions but less"
                " than or equal to 4"
            )
            if attn_mask.ndim == 2:
                attn_mask = rearrange(attn_mask, "i j -> 1 1 i j")
            elif attn_mask.ndim == 3:
                attn_mask = rearrange(attn_mask, "h i j -> 1 h i j")
            masks.append(~attn_mask)

        if exists(self.max_attend_past):
            range_q = torch.arange(j - i, j, device=device)
            range_k = torch.arange(j, device=device)
            dist = rearrange(range_q, "i -> 1 1 i 1") - rearrange(
                range_k, "j -> 1 1 1 j"
            )
            max_attend_past_mask = dist > self.max_attend_past
            masks.append(max_attend_past_mask)

        if len(masks) > 0:
            final_attn_mask = ~or_reduce(masks)

        # prepare relative positional bias, if needed

        attn_bias = None
        if exists(rel_pos):
            attn_bias = rel_pos(i, j)

        # attention is all we need

        out, intermediates = self.attend(
            q,
            k,
            v,
            mask=final_attn_mask,
            attn_bias=attn_bias,
            prev_attn=prev_attn,
        )

        # https://arxiv.org/abs/2208.06061 proposes to add a residual for
        # better gradients

        if exists(r):
            out = out * r + out

        # normformer scaling of heads

        if head_scale:
            out = out * self.head_scale_params

        # merge heads

        out = rearrange(out, "b h n d -> b n (h d)")

        # alphafold2 styled gating of the values

        if exists(self.to_v_gate):
            gates = self.to_v_gate(x)
            out = out * gates.sigmoid()

        # combine the heads

        out = self.to_out(out)

        if exists(mask):
            mask = rearrange(mask, "b n -> b n 1")
            out = out.masked_fill(~mask, 0.0)

        return out, intermediates


class AttentionLayers(nn.Module):
    def __init__(
        self,
        dim,
        depth,
        heads=8,
        causal=False,
        cross_attend=False,
        only_cross=False,
        use_scalenorm=False,
        use_rmsnorm=False,
        use_simple_rmsnorm=False,
        alibi_pos_bias=False,
        alibi_num_heads=None,
        rel_pos_bias=False,
        rel_pos_num_buckets=32,
        rel_pos_max_distance=128,
        dynamic_pos_bias=False,
        dynamic_pos_bias_log_distance=False,
        dynamic_pos_bias_mlp_depth=2,
        dynamic_pos_bias_norm=False,
        rotary_pos_emb=False,
        rotary_emb_dim=None,
        rotary_xpos=False,
        rotary_interpolation_factor=1.0,
        rotary_xpos_scale_base=512,
        rotary_base_rescale_factor=1.0,
        custom_layers=None,
        sandwich_coef=None,
        par_ratio=None,
        residual_attn=False,
        cross_residual_attn=False,
        macaron=False,
        pre_norm=True,
        pre_norm_has_final_norm=True,
        gate_residual=False,
        scale_residual=False,
        scale_residual_constant=1.0,
        deepnorm=False,
        shift_tokens=0,
        sandwich_norm=False,
        resi_dual=False,
        resi_dual_scale=1.0,
        zero_init_branch_output=False,
        layer_dropout=0.0,
        cross_attn_tokens_dropout=0.0,
        **kwargs,
    ):
        super().__init__()
        rotary_pos_emb = rotary_pos_emb or rotary_xpos

        ff_kwargs, kwargs = groupby_prefix_and_trim("ff_", kwargs)
        attn_kwargs, kwargs = groupby_prefix_and_trim("attn_", kwargs)

        dim_head = attn_kwargs.get("dim_head", DEFAULT_DIM_HEAD)

        self.dim = dim
        self.depth = depth
        self.layers = nn.ModuleList([])

        self.has_pos_emb = rel_pos_bias or rotary_pos_emb

        rotary_emb_dim = max(default(rotary_emb_dim, dim_head // 2), 32)

        assert not (
            rotary_xpos and not causal
        ), "rotary xpos is not compatible with bidirectional attention"
        self.rotary_pos_emb = (
            RotaryEmbedding(
                rotary_emb_dim,
                use_xpos=rotary_xpos,
                scale_base=rotary_xpos_scale_base,
                interpolation_factor=rotary_interpolation_factor,
                base_rescale_factor=rotary_base_rescale_factor,
            )
            if rotary_pos_emb
            else None
        )

        assert not (alibi_pos_bias and rel_pos_bias), (
            "you can only choose Alibi positional bias or T5 relative"
            " positional bias, not both"
        )
        assert rel_pos_num_buckets <= rel_pos_max_distance, (
            "number of relative position buckets must be less than the relative"
            " position max distance"
        )

        # relative positional bias

        flash_attn = attn_kwargs.get("flash", False)
        assert (
            int(rel_pos_bias) + int(dynamic_pos_bias) + int(alibi_pos_bias)
        ) <= 1, (
            "you can only choose up to one of t5, alibi, or dynamic positional"
            " bias"
        )

        self.rel_pos = None
        if rel_pos_bias:
            assert (
                not flash_attn
            ), "flash attention not compatible with t5 relative positional bias"
            self.rel_pos = RelativePositionBias(
                scale=dim_head**0.5,
                causal=causal,
                heads=heads,
                num_buckets=rel_pos_num_buckets,
                max_distance=rel_pos_max_distance,
            )
        elif dynamic_pos_bias:
            assert (
                not flash_attn
            ), "flash attention not compatible with dynamic positional bias"
            self.rel_pos = DynamicPositionBias(
                dim=dim // 4,
                heads=heads,
                log_distance=dynamic_pos_bias_log_distance,
                depth=dynamic_pos_bias_mlp_depth,
                norm=dynamic_pos_bias_norm,
            )
        elif alibi_pos_bias:
            alibi_num_heads = default(alibi_num_heads, heads)
            assert alibi_num_heads <= heads, (
                "number of ALiBi heads must be less than the total number of"
                " heads"
            )
            self.rel_pos = AlibiPositionalBias(
                heads=alibi_num_heads, total_heads=heads
            )

        # determine deepnorm and residual scale

        if deepnorm:
            assert scale_residual_constant == 1, (
                "scale residual constant is being overridden by deep norm"
                " settings"
            )
            pre_norm = sandwich_norm = resi_dual = False
            scale_residual = True
            scale_residual_constant = (2 * depth) ** 0.25

        assert (
            int(sandwich_norm) + int(resi_dual)
        ) <= 1, "either sandwich norm or resiDual is selected, but not both"
        assert not (
            not pre_norm and sandwich_norm
        ), "sandwich norm cannot be used when not using prenorm"

        if resi_dual:
            pre_norm = False

        self.pre_norm = pre_norm
        self.sandwich_norm = sandwich_norm

        self.resi_dual = resi_dual
        assert 0 < resi_dual_scale <= 1.0, (
            "resiDual prenorm residual must be scaled by a factor greater than"
            " 0 and less than or equal to 1."
        )
        self.resi_dual_scale = resi_dual_scale

        self.residual_attn = residual_attn
        self.cross_residual_attn = cross_residual_attn
        assert not (
            flash_attn and (residual_attn or cross_residual_attn)
        ), "flash attention is not compatible with residual attention"

        self.cross_attend = cross_attend

        assert (
            int(use_scalenorm) + int(use_rmsnorm) + int(use_simple_rmsnorm)
        ) <= 1, "you can only use either scalenorm, rmsnorm, or simple rmsnorm"

        if use_scalenorm:
            norm_class = ScaleNorm
        elif use_rmsnorm:
            norm_class = RMSNorm
        elif use_simple_rmsnorm:
            norm_class = SimpleRMSNorm
        else:
            norm_class = nn.LayerNorm

        norm_fn = partial(norm_class, dim)

        if cross_attend and not only_cross:
            default_block = ("a", "c", "f")
        elif cross_attend and only_cross:
            default_block = ("c", "f")
        else:
            default_block = ("a", "f")

        if macaron:
            default_block = ("f",) + default_block

        # zero init

        if zero_init_branch_output:
            attn_kwargs = {**attn_kwargs, "zero_init_output": True}
            ff_kwargs = {**ff_kwargs, "zero_init_output": True}

        # calculate layer block order

        if exists(custom_layers):
            layer_types = custom_layers
        elif exists(par_ratio):
            par_depth = depth * len(default_block)
            assert 1 < par_ratio <= par_depth, "par ratio out of range"
            default_block = tuple(filter(not_equals("f"), default_block))
            par_attn = par_depth // par_ratio
            # 2 / 3 attention layer cutoff suggested by PAR paper
            depth_cut = par_depth * 2 // 3
            par_width = (depth_cut + depth_cut // par_attn) // par_attn
            assert (
                len(default_block) <= par_width
            ), "default block is too large for par_ratio"
            par_block = default_block + ("f",) * (
                par_width - len(default_block)
            )
            par_head = par_block * par_attn
            layer_types = par_head + ("f",) * (par_depth - len(par_head))
        elif exists(sandwich_coef):
            assert (
                sandwich_coef > 0 and sandwich_coef <= depth
            ), "sandwich coefficient should be less than the depth"
            layer_types = (
                ("a",) * sandwich_coef
                + default_block * (depth - sandwich_coef)
                + ("f",) * sandwich_coef
            )
        else:
            layer_types = default_block * depth

        self.layer_types = layer_types
        self.num_attn_layers = len(list(filter(equals("a"), layer_types)))

        # stochastic depth

        self.layer_dropouts = cast_tuple(layer_dropout, len(layer_types))

        # structured dropout for cross attending

        self.cross_attn_tokens_dropout = cross_attn_tokens_dropout

        # calculate token shifting

        shift_tokens = cast_tuple(shift_tokens, len(layer_types))

        # whether it has post norm

        self.final_norm = norm_fn() if pre_norm or resi_dual else nn.Identity()

        # iterate and construct layers

        for ind, (layer_type, layer_shift_tokens) in enumerate(
            zip(self.layer_types, shift_tokens)
        ):
            ind == (len(self.layer_types) - 1)

            if layer_type == "a":
                layer = Attention(
                    dim, heads=heads, causal=causal, **attn_kwargs
                )
            elif layer_type == "c":
                layer = Attention(dim, heads=heads, **attn_kwargs)
            elif layer_type == "f":
                layer = FeedForward(dim, **ff_kwargs)
                layer = layer if not macaron else Scale(0.5, layer)
            else:
                raise Exception(f"invalid layer type {layer_type}")

            if layer_shift_tokens > 0:
                shift_range_upper = layer_shift_tokens + 1
                shift_range_lower = -layer_shift_tokens if not causal else 0
                layer = ShiftTokens(
                    range(shift_range_lower, shift_range_upper), layer
                )

            residual_fn = GRUGating if gate_residual else Residual
            residual = residual_fn(
                dim,
                scale_residual=scale_residual,
                scale_residual_constant=scale_residual_constant,
            )

            pre_branch_norm = norm_fn() if pre_norm else None
            post_branch_norm = norm_fn() if sandwich_norm else None
            post_main_norm = norm_fn() if not pre_norm else None

            norms = nn.ModuleList(
                [pre_branch_norm, post_branch_norm, post_main_norm]
            )

            self.layers.append(nn.ModuleList([norms, layer, residual]))

        if deepnorm:
            init_gain = (8 * depth) ** -0.25
            deepnorm_init(self, init_gain)

    def forward(
        self,
        x,
        context=None,
        mask=None,
        context_mask=None,
        attn_mask=None,
        self_attn_context_mask=None,
        mems=None,
        return_hiddens=False,
    ):
        assert not (
            self.cross_attend ^ exists(context)
        ), "context must be passed in if cross_attend is set to True"

        hiddens = []
        layer_hiddens = []
        intermediates = []

        prev_attn = None
        prev_cross_attn = None

        mems = mems.copy() if exists(mems) else [None] * self.num_attn_layers

        rotary_pos_emb = None
        if exists(self.rotary_pos_emb):
            max_rotary_emb_length = max(
                list(
                    map(
                        lambda m: (m.shape[1] if exists(m) else 0) + x.shape[1],
                        mems,
                    )
                )
            )
            rotary_pos_emb = self.rotary_pos_emb(
                max_rotary_emb_length, x.device
            )

        outer_residual = x * self.resi_dual_scale

        for ind, (
            layer_type,
            (norm, block, residual_fn),
            layer_dropout,
        ) in enumerate(zip(self.layer_types, self.layers, self.layer_dropouts)):
            ind == (len(self.layers) - 1)

            if (
                self.training
                and layer_dropout > 0.0
                and random() < layer_dropout
            ):
                continue

            if layer_type == "a":
                if return_hiddens:
                    hiddens.append(x)
                layer_mem = mems.pop(0) if mems else None

            if layer_type == "c":
                if self.training and self.cross_attn_tokens_dropout > 0.0:
                    context, context_mask = dropout_seq(
                        context, context_mask, self.cross_attn_tokens_dropout
                    )

            inner_residual = x

            if return_hiddens:
                layer_hiddens.append(x)

            pre_norm, post_branch_norm, post_main_norm = norm

            if exists(pre_norm):
                x = pre_norm(x)

            if layer_type == "a":
                out, inter = block(
                    x,
                    mask=mask,
                    context_mask=self_attn_context_mask,
                    attn_mask=attn_mask,
                    rel_pos=self.rel_pos,
                    rotary_pos_emb=rotary_pos_emb,
                    prev_attn=prev_attn,
                    mem=layer_mem,
                )
            elif layer_type == "c":
                out, inter = block(
                    x,
                    context=context,
                    mask=mask,
                    context_mask=context_mask,
                    prev_attn=prev_cross_attn,
                )
            elif layer_type == "f":
                out = block(x)

            if self.resi_dual:
                outer_residual = outer_residual + out * self.resi_dual_scale

            if exists(post_branch_norm):
                out = post_branch_norm(out)

            x = residual_fn(out, inner_residual)

            if layer_type in ("a", "c") and return_hiddens:
                intermediates.append(inter)

            if layer_type == "a" and self.residual_attn:
                prev_attn = inter.pre_softmax_attn
            elif layer_type == "c" and self.cross_residual_attn:
                prev_cross_attn = inter.pre_softmax_attn

            if exists(post_main_norm):
                x = post_main_norm(x)

        if return_hiddens:
            layer_hiddens.append(x)

        if self.resi_dual:
            x = x + self.final_norm(outer_residual)
        else:
            x = self.final_norm(x)

        if return_hiddens:
            intermediates = LayerIntermediates(
                hiddens=hiddens,
                attn_intermediates=intermediates,
                layer_hiddens=layer_hiddens,
            )

            return x, intermediates

        return x


class Encoder(AttentionLayers):
    def __init__(self, **kwargs):
        assert "causal" not in kwargs, "cannot set causality on encoder"
        super().__init__(causal=False, **kwargs)


class Decoder(AttentionLayers):
    def __init__(self, **kwargs):
        assert "causal" not in kwargs, "cannot set causality on decoder"
        super().__init__(causal=True, **kwargs)


class CrossAttender(AttentionLayers):
    def __init__(self, **kwargs):
        super().__init__(cross_attend=True, only_cross=True, **kwargs)


class ViTransformerWrapper(nn.Module):
    def __init__(
        self,
        *,
        image_size,
        patch_size,
        attn_layers,
        channels=3,
        num_classes=None,
        post_emb_norm=False,
        emb_dropout=0.0,
    ):
        super().__init__()
        assert isinstance(
            attn_layers, Encoder
        ), "attention layers must be an Encoder"
        assert divisible_by(
            image_size, patch_size
        ), "image dimensions must be divisible by the patch size"
        dim = attn_layers.dim
        num_patches = (image_size // patch_size) ** 2
        patch_dim = channels * patch_size**2

        self.patch_size = patch_size

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches, dim))

        self.patch_to_embedding = nn.Sequential(
            nn.LayerNorm(patch_dim),
            nn.Linear(patch_dim, dim),
            nn.LayerNorm(dim),
        )

        self.post_emb_norm = (
            nn.LayerNorm(dim) if post_emb_norm else nn.Identity()
        )
        self.dropout = nn.Dropout(emb_dropout)

        self.attn_layers = attn_layers

        self.mlp_head = (
            nn.Linear(dim, num_classes)
            if exists(num_classes)
            else nn.Identity()
        )

    def forward(self, img, return_embeddings=False):
        p = self.patch_size

        x = rearrange(img, "b c (h p1) (w p2) -> b (h w) (p1 p2 c)", p1=p, p2=p)
        x = self.patch_to_embedding(x)
        n = x.shape[1]

        x = x + self.pos_embedding[:, :n]

        x = self.post_emb_norm(x)
        x = self.dropout(x)

        x = self.attn_layers(x)

        if not exists(self.mlp_head) or return_embeddings:
            return x

        x = x.mean(dim=-2)
        return self.mlp_head(x)


class Transformer(nn.Module):
    def __init__(
        self,
        *,
        num_tokens,
        max_seq_len,
        attn_layers,
        emb_dim=None,
        max_mem_len=0,
        shift_mem_down=0,
        emb_dropout=0.0,
        post_emb_norm=False,
        num_memory_tokens=None,
        tie_embedding=False,
        logits_dim=None,
        use_abs_pos_emb=True,
        scaled_sinu_pos_emb=False,
        l2norm_embed=False,
        emb_frac_gradient=1.0,  # GLM-130B and Cogview successfully used this, set at 0.1
        attn_z_loss_weight=1e-4,
    ):
        super().__init__()
        assert isinstance(
            attn_layers, AttentionLayers
        ), "attention layers must be one of Encoder or Decoder"

        dim = attn_layers.dim
        emb_dim = default(emb_dim, dim)
        self.emb_dim = emb_dim
        self.num_tokens = num_tokens

        self.max_seq_len = max_seq_len
        self.max_mem_len = max_mem_len
        self.shift_mem_down = shift_mem_down

        self.l2norm_embed = l2norm_embed
        self.token_emb = TokenEmbedding(
            emb_dim, num_tokens, l2norm_embed=l2norm_embed
        )

        if not (use_abs_pos_emb and not attn_layers.has_pos_emb):
            self.pos_emb = always(0)
        elif scaled_sinu_pos_emb:
            self.pos_emb = ScaledSinusoidalEmbedding(emb_dim)
        else:
            self.pos_emb = AbsolutePositionalEmbedding(
                emb_dim, max_seq_len, l2norm_embed=l2norm_embed
            )

        # fraction of the gradient that should go to the embedding,
        # https://arxiv.org/abs/2105.13290
        self.emb_frac_gradient = emb_frac_gradient

        self.post_emb_norm = (
            nn.LayerNorm(emb_dim) if post_emb_norm else nn.Identity()
        )
        self.emb_dropout = nn.Dropout(emb_dropout)

        self.project_emb = (
            nn.Linear(emb_dim, dim) if emb_dim != dim else nn.Identity()
        )
        self.attn_layers = attn_layers

        self.init_()

        logits_dim = default(logits_dim, num_tokens)
        self.to_logits = (
            nn.Linear(dim, logits_dim)
            if not tie_embedding
            else lambda t: t @ self.token_emb.emb.weight.t()
        )

        # memory tokens (like [cls]) from Memory Transformers paper
        num_memory_tokens = default(num_memory_tokens, 0)
        self.num_memory_tokens = num_memory_tokens
        if num_memory_tokens > 0:
            self.memory_tokens = nn.Parameter(
                torch.randn(num_memory_tokens, dim)
            )

    def init_(self):
        if self.l2norm_embed:
            nn.init.normal_(self.token_emb.emb.weight, std=1e-5)
            if not isinstance(self.pos_emb, always):
                nn.init.normal_(self.pos_emb.emb.weight, std=1e-5)
            return

        nn.init.kaiming_normal_(self.token_emb.emb.weight)

    def forward(
        self,
        x,
        return_embeddings=False,
        return_logits_and_embeddings=False,
        return_intermediates=False,
        mask=None,
        return_mems=False,
        return_attn=False,
        mems=None,
        pos=None,
        prepend_embeds=None,
        sum_embeds=None,
        return_attn_z_loss=False,
        attn_z_loss_weight=1e-4,
        **kwargs,
    ):
        # einops conflicts with ruff, so noqa on next line
        b, n, device, num_mem, emb_frac_gradient = (  # noqa F841
            *x.shape,
            x.device,
            self.num_memory_tokens,
            self.emb_frac_gradient,
        )
        return_hiddens = (
            return_mems
            | return_attn
            | return_intermediates
            | return_attn_z_loss
        )

        # absolute positional embedding

        external_pos_emb = exists(pos) and pos.dtype != torch.long
        pos_emb = self.pos_emb(x, pos=pos) if not external_pos_emb else pos
        x = self.token_emb(x) + pos_emb

        # for summing embeddings passed externally - needs this for
        # self-conditioning in non-autoregressive training

        if exists(sum_embeds):
            x = x + sum_embeds

        # post embedding norm, purportedly leads to greater stabilization

        x = self.post_emb_norm(x)

        # whether to append embeds, as in PaLI, for image embeddings

        if exists(prepend_embeds):
            prepend_seq, prepend_dim = prepend_embeds.shape[1:]
            assert prepend_dim == x.shape[-1], (
                "prepended embeddings need to have same dimensions as text"
                " model dimensions"
            )

            x = torch.cat((prepend_embeds, x), dim=-2)

        # whether to reduce the gradient going to the embedding, from cogview
        # paper, corroborated by GLM-130B model

        if emb_frac_gradient < 1:
            assert emb_frac_gradient > 0
            x = x * emb_frac_gradient + x.detach() * (1 - emb_frac_gradient)

        # embedding dropout

        x = self.emb_dropout(x)

        x = self.project_emb(x)

        if num_mem > 0:
            mem = repeat(self.memory_tokens, "n d -> b n d", b=b)
            x = torch.cat((mem, x), dim=1)

            # auto-handle masking after appending memory tokens
            if exists(mask):
                mask = pad_at_dim(mask, (num_mem, 0), dim=-1, value=True)

        if self.shift_mem_down and exists(mems):
            mems_l, mems_r = (
                mems[: self.shift_mem_down],
                mems[self.shift_mem_down :],
            )
            mems = [*mems_r, *mems_l]

        if return_hiddens:
            x, intermediates = self.attn_layers(
                x, mask=mask, mems=mems, return_hiddens=True, **kwargs
            )
        else:
            x = self.attn_layers(x, mask=mask, mems=mems, **kwargs)

        mem, x = x[:, :num_mem], x[:, num_mem:]

        if return_logits_and_embeddings:
            out = (self.to_logits(x), x)
        elif return_embeddings:
            out = x
        else:
            out = self.to_logits(x)

        if return_attn_z_loss:
            pre_softmax_attns = list(
                map(
                    lambda t: t.pre_softmax_attn,
                    intermediates.attn_intermediates,
                )
            )
            intermediates.attn_z_loss = calc_z_loss(
                pre_softmax_attns, weight=attn_z_loss_weight
            )
            return_intermediates = True

        if return_intermediates:
            return out, intermediates

        if return_mems:
            hiddens = intermediates.hiddens
            new_mems = (
                list(
                    map(
                        lambda pair: torch.cat(pair, dim=-2), zip(mems, hiddens)
                    )
                )
                if exists(mems)
                else hiddens
            )
            new_mems = list(
                map(lambda t: t[..., -self.max_mem_len :, :].detach(), new_mems)
            )
            return out, new_mems

        if return_attn:
            attn_maps = list(
                map(
                    lambda t: t.post_softmax_attn,
                    intermediates.attn_intermediates,
                )
            )
            return out, attn_maps

        return out



import torch
import torch.nn as nn


class MultiLayerKeyValueAttention(nn.Module):
    """
    Multi-layer key-value attention module.

    Args:
        embed_size (int): The size of the input embeddings.
        num_heads (int): The number of attention heads.
        num_layers (int): The number of layers.
        kv_layers (int): The number of key-value layers.

    Attributes:
        num_heads (int): The number of attention heads.
        num_layers (int): The number of layers.
        kv_layers (int): The number of key-value layers.
        embed_size (int): The size of the input embeddings.
        head_dim (int): The dimension of each attention head.

        values (nn.ModuleList): List of value projection layers for each key-value layer.
        keys (nn.ModuleList): List of key projection layers for each key-value layer.
        queries (nn.ModuleList): List of query projection layers for each layer.
        fc_out (nn.Linear): Output linear layer.

    """

    def __init__(self, embed_size, num_heads, num_layers, kv_layers):
        super(MultiLayerKeyValueAttention, self).__init__()
        self.num_heads = num_heads
        self.num_layers = num_layers
        self.kv_layers = kv_layers  # m in the description
        self.embed_size = embed_size
        self.head_dim = embed_size // num_heads

        assert (
            self.head_dim * num_heads == embed_size
        ), "Embedding size needs to be divisible by num_heads"

        # Define the key and value projections for each layer
        self.values = nn.ModuleList(
            [
                nn.Linear(embed_size, embed_size, bias=False)
                for _ in range(kv_layers)
            ]
        )
        self.keys = nn.ModuleList(
            [
                nn.Linear(embed_size, embed_size, bias=False)
                for _ in range(kv_layers)
            ]
        )

        # Define the query projections for each layer
        self.queries = nn.ModuleList(
            [
                nn.Linear(embed_size, embed_size, bias=False)
                for _ in range(num_layers)
            ]
        )

        self.fc_out = nn.Linear(embed_size, embed_size)

    def forward(self, values, keys, queries):
        """
        Forward pass of the multi-layer key-value attention module.

        Args:
            values (torch.Tensor): The values tensor of shape (N, value_len, embed_size).
            keys (torch.Tensor): The keys tensor of shape (N, key_len, embed_size).
            queries (torch.Tensor): The queries tensor of shape (N, query_len, embed_size).

        Returns:
            torch.Tensor: The output tensor of shape (N, query_len, embed_size).

        """
        N = queries.shape[0]
        value_len, key_len, query_len = (
            values.shape[1],
            keys.shape[1],
            queries.shape[1],
        )

        out = torch.zeros(N, query_len, self.embed_size).to(values.device)

        for layer in range(self.num_layers):
            kv_index = layer % self.kv_layers

            values_layer = self.values[kv_index](values).view(
                N, value_len, self.num_heads, self.head_dim
            )
            keys_layer = self.keys[kv_index](keys).view(
                N, key_len, self.num_heads, self.head_dim
            )
            queries_layer = self.queries[layer](queries).view(
                N, query_len, self.num_heads, self.head_dim
            )

            energy = torch.einsum(
                "nqhd,nkhd->nhqk", [queries_layer, keys_layer]
            )
            attention = torch.softmax(
                energy / (self.embed_size ** (1 / 2)), dim=3
            )
            out_layer = torch.einsum(
                "nhql,nlhd->nqhd", [attention, values_layer]
            ).reshape(N, query_len, self.embed_size)

            out += out_layer

        out = self.fc_out(out)
        return out


# # Example usage
# embed_size = 256
# num_heads = 8
# num_layers = 4
# kv_layers = 2  # Number of layers with their own KV heads

# mlkv_attention = MultiLayerKeyValueAttention(embed_size, num_heads, num_layers, kv_layers)
# values = torch.rand(32, 10, embed_size)  # batch size 32, sequence length 10
# keys = torch.rand(32, 10, embed_size)
# queries = torch.rand(32, 10, embed_size)

# output = mlkv_attention(values, keys, queries)
# print(output.shape)



from torch import nn


class ImgPatchEmbed(nn.Module):
    """patch embedding module


    Args:
        img_size (int, optional): image size. Defaults to 224.
        patch_size (int, optional): patch size. Defaults to 16.
        in_chans (int, optional): input channels. Defaults to 3.
        embed_dim (int, optional): embedding dimension. Defaults to 768.

    Examples:
        >>> x = torch.randn(1, 3, 224, 224)
        >>> model = ImgPatchEmbed()
        >>> model(x).shape
        torch.Size([1, 196, 768])


    """

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        num_patches = (img_size // patch_size) * (img_size // patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )

    def forward(self, x):
        """Forward

        Args:
            x (_type_): _description_

        Returns:
            _type_: _description_
        """
        B, C, H, W = x.shape
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x



import torch
import torch.nn.functional as F
from torch import nn


class LayerNorm(nn.Module):
    """
    Layer normalization module.

    Args:
        dim (int): The dimension of the input tensor.
        eps (float, optional): A small value added to the denominator for numerical stability. Default is 1e-5.
        fp16_eps (float, optional): A small value added to the denominator for numerical stability when using fp16 data type. Default is 1e-3.
        stable (bool, optional): Whether to use a stable implementation of layer normalization. Default is False.

    Attributes:
        eps (float): A small value added to the denominator for numerical stability.
        fp16_eps (float): A small value added to the denominator for numerical stability when using fp16 data type.
        stable (bool): Whether to use a stable implementation of layer normalization.
        g (torch.nn.Parameter): The learnable scale parameter.

    # Usage
    ```
    import torch

    # Create an instance of LayerNorm
    layer_norm = LayerNorm(dim=10)

    # Create a random input tensor
    x = torch.randn(32, 10)

    # Apply layer normalization
    normalized_x = layer_norm(x)

    # Print the normalized tensor
    print(normalized_x)

    # Apply L2 normalization
    normalized_x = l2norm(x)

    # Print the normalized tensor
    print(normalized_x)
    ```

    """

    def __init__(self, dim, eps=1e-5, fp16_eps=1e-3, stable=False):
        super().__init__()
        self.eps = eps
        self.fp16_eps = fp16_eps
        self.stable = stable
        self.g = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        """
        Forward pass of the layer normalization module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The normalized tensor.

        """
        eps = self.eps if x.dtype == torch.float32 else self.fp16_eps

        if self.stable:
            x = x / x.amax(dim=-1, keepdim=True).detach()

        var = torch.var(x, dim=-1, unbiased=False, keepdim=True)
        mean = torch.mean(x, dim=-1, keepdim=True)
        return (x - mean) * (var + eps).rsqrt() * self.g


def l2norm(t):
    """
    L2 normalization function.

    Args:
        t (torch.Tensor): The input tensor.

    Returns:
        torch.Tensor: The normalized tensor.

    """
    return F.normalize(t, dim=-1)



from typing import Callable, List, Optional, Tuple

from beartype import beartype
from einops.layers.torch import Rearrange, Reduce
from torch import nn

from zeta.nn.attention.attend import Attend
from zeta.nn.modules.layernorm import LayerNorm
from zeta.nn.modules.mbconv import MBConv
from zeta.structs.transformer import FeedForward, Residual
from zeta.utils.main import default, exists


class MaxVit(nn.Module):
    def __init__(
        self,
        *,
        num_classes,
        dim,
        depth,
        dim_head: int = 32,
        dim_conv_stem=None,
        window_size: int = 7,
        mbconv_expansion_rate: int = 4,
        mbconv_shrinkage_rate=0.25,
        dropout=0.01,
        channels=3,
    ):
        super().__init__()
        assert isinstance(depth, tuple), (
            "depth needs to be tuple of integers indicating number of"
            " transformer blocks at that stage"
        )

        # conv stem
        dim_conv_stem = default(dim_conv_stem, dim)

        self.conv_stem = nn.Sequential(
            nn.Conv2d(
                channels,
                dim_conv_stem,
                3,
                stride=2,
                padding=1,
            ),
            nn.Conv2d(dim_conv_stem, dim_conv_stem, 3, padding=1),
        )

        # vars
        num_stages = len(depth)
        dims = tuple(map(lambda i: (2**i) * dim, range(num_stages)))
        dims = (dim_conv_stem, *dims)
        dim_pairs = tuple(zip(dims[:-1], dims[1:]))

        self.layers = nn.ModuleList([])

        # shorthand for window size for efficient block - griid like attention
        w = window_size

        # iterate through stages
        cond_hidden_dims = []

        for ind, ((layer_dim_in, layer_dim), layer_depth) in enumerate(
            zip(dim_pairs, depth)
        ):
            for stage_ind in range(layer_depth):
                is_first = stage_ind == 0
                stage_dim_in = layer_dim_in if is_first else layer_dim

                cond_hidden_dims.append(stage_dim_in)

                block = nn.Sequential(
                    MBConv(
                        stage_dim_in,
                        layer_dim,
                        downsample=is_first,
                        expansion_rate=mbconv_expansion_rate,
                        shrinkage_rate=mbconv_shrinkage_rate,
                    ),
                    Rearrange("b d (x w1) (y w2) -> b x y w1 w2 d", w1=w, w2=w),
                    Residual(
                        Attend(
                            dim=layer_dim, dim_head=dim_head, dropout=dropout
                        )
                    ),
                    Residual(FeedForward(dim=layer_dim, dropout=dropout)),
                    Rearrange("b x y w1 w2 d -> b d (w1 x) (w2 y)"),
                )

                self.layers.append(block)

        embed_dim = dims[-1]
        self.embed_dim = dims[-1]

        self.mlp_head = nn.Sequential(
            Reduce("b d h w -> b d", "mean"),
            LayerNorm,
            nn.Linear(embed_dim, num_classes),
        )

    @beartype
    def forward(
        self,
        x,
        texts: Optional[List[str]] = None,
        cond_fns: Optional[Tuple[Callable, ...]] = None,
        cond_drop_prob=0.0,
        return_embeddings=False,
    ):
        x = self.conv_stem(x)

        if not exists(cond_fns):
            cond_fns = (None,) * len(self.layers)

        for stage, cond_fn in zip(self.layers, cond_fns):
            if exists(cond_fn):
                x = cond_fn(x)

            x = stage(x)

        if return_embeddings:
            return x

        return self.mlp_head(x)



import torch
from torch import nn, Tensor


class ScaleNorm(nn.Module):
    """
    Applies scale normalization to the input tensor along the last dimension.

    Args:
        dim (int): The dimension of the input tensor.
        eps (float, optional): A small value added to the denominator for numerical stability. Default is 1e-5.
    """

    def __init__(
        self,
        dim: int,
        eps: float = 1e-5,
    ):
        super().__init__()
        self.eps = eps

        self.g = nn.Parameter(torch.ones(1) * (dim**-0.5))

    def forward(self, x: Tensor):
        """
        Applies scale normalization to the input tensor.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The scale-normalized tensor.
        """
        norm = torch.norm(x, dim=-1, keepdim=True)
        return x / norm.clamp(min=self.eps) + self.g



# from zeta.tokenizers.gptx_tokenizer import LanguageTokenizerGPTX
# from zeta.tokenizers.llama_sentencepiece import LLamaTokenizer
# from zeta.tokenizers.multi_modal_tokenizer import MultiModalTokenizer
# from zeta.tokenizers.sentence_piece import SentencePieceTokenizer
# from zeta.tokenizers.tokenmonster import TokenMonster

# __all__ = [
#     "LanguageTokenizerGPTX",
#     "MultiModalTokenizer",
#     "SentencePieceTokenizer",
#     "TokenMonster",
#     "LLamaTokenizer",
# ]



from zeta.nn.modules._activations import (
    AccurateGELUActivation,
    ClippedGELUActivation,
    FastGELUActivation,
    GELUActivation,
    LaplaceActivation,
    LinearActivation,
    MishActivation,
    NewGELUActivation,
    PytorchGELUTanh,
    QuickGELUActivation,
    ReLUSquaredActivation,
)
from zeta.nn.modules.adaptive_conv import AdaptiveConv3DMod
from zeta.nn.modules.adaptive_layernorm import AdaptiveLayerNorm
from zeta.nn.modules.adaptive_rmsnorm import AdaptiveRMSNorm
from zeta.nn.modules.add_norm import add_norm
from zeta.nn.modules.audio_to_text import audio_to_text
from zeta.nn.modules.avg_model_merger import AverageModelMerger
from zeta.nn.modules.block_butterfly_mlp import BlockButterflyLinear, BlockMLP
from zeta.nn.modules.blockdiag_butterfly import (
    BlockdiagButterflyMultiply,
    BlockdiagMultiply,
    Sin,
    StructuredLinear,
    blockdiag_butterfly_multiply_reference,
    blockdiag_multiply_reference,
    blockdiag_weight_to_dense_weight,
    fftconv_ref,
    mul_sum,
)
from zeta.nn.modules.cnn_text import CNNNew
from zeta.nn.modules.combined_linear import CombinedLinear
from zeta.nn.modules.conv_mlp import Conv2DFeedforward
from zeta.nn.modules.convnet import ConvNet
from zeta.nn.modules.cross_modal_reparametization import (
    CrossModalReParametrization,
    CrossModalReparamLinear,
    build_cross_modal_reparam_linear,
    change_original_linear_to_reparam,
    cross_modal_ffn,
    reparameterize_aux_into_target_model,
)
from zeta.nn.modules.dense_connect import DenseBlock
from zeta.nn.modules.dual_path_block import DualPathBlock
from zeta.nn.modules.dynamic_module import DynamicModule
from zeta.nn.modules.dynamic_routing_block import DynamicRoutingBlock
from zeta.nn.modules.ether import Ether
from zeta.nn.modules.exo import Exo
from zeta.nn.modules.fast_text import FastTextNew
from zeta.nn.modules.feedback_block import FeedbackBlock
from zeta.nn.modules.feedforward import FeedForward
from zeta.nn.modules.feedforward_network import FeedForwardNetwork
from zeta.nn.modules.film import Film
from zeta.nn.modules.film_conditioning import FilmConditioning
from zeta.nn.modules.flex_conv import FlexiConv
from zeta.nn.modules.flexible_mlp import CustomMLP
from zeta.nn.modules.freeze_layers import (
    freeze_all_layers,
    set_module_requires_grad,
)
from zeta.nn.modules.fused_dropout_add import (
    fused_bias_dropout_add,
    fused_dropout_add,
    jit_bias_dropout_add,
    jit_dropout_add,
)
from zeta.nn.modules.fused_dropout_layernom import FusedDropoutLayerNorm
from zeta.nn.modules.fused_gelu_dense import FusedDenseGELUDense
from zeta.nn.modules.fusion_ffn import MMFusionFFN
from zeta.nn.modules.gated_residual_block import GatedResidualBlock
from zeta.nn.modules.gill_mapper import GILLMapper
from zeta.nn.modules.h3 import H3Layer
from zeta.nn.modules.highway_layer import HighwayLayer
from zeta.nn.modules.image_to_text import img_to_text
from zeta.nn.modules.img_or_video_to_time import image_or_video_to_time
from zeta.nn.modules.img_patch_embed import ImgPatchEmbed
from zeta.nn.modules.itca import IterativeCrossSelfAttention
from zeta.nn.modules.lang_conv_module import ConvolutionLanguageBlock
from zeta.nn.modules.laser import Laser
from zeta.nn.modules.layernorm import LayerNorm, l2norm
from zeta.nn.modules.leaky_relu import LeakyRELU
from zeta.nn.modules.log_ff import LogFF
from zeta.nn.modules.lora import Lora
from zeta.nn.modules.mbconv import (
    DropSample,
    MBConv,
    MBConvResidual,
    SqueezeExcitation,
)
from zeta.nn.modules.mlp import MLP
from zeta.nn.modules.mlp_mixer import MixerBlock, MLPBlock, MLPMixer
from zeta.nn.modules.mm_layernorm import MMLayerNorm
from zeta.nn.modules.mm_ops import text_to_twod, threed_to_text
from zeta.nn.modules.moe import MixtureOfExperts
from zeta.nn.modules.moe_router import MoERouter
from zeta.nn.modules.multi_input_multi_output import (
    DynamicInputChannels,
    DynamicOutputDecoder,
    MultiInputMultiModalConcatenation,
    MultiModalEmbedding,
    OutputDecoders,
    OutputHead,
    SplitMultiOutput,
)
from zeta.nn.modules.multi_scale_block import MultiScaleBlock
from zeta.nn.modules.nebula import Nebula
from zeta.nn.modules.nfn_stem import NFNStem
from zeta.nn.modules.norm_fractorals import NormalizationFractral
from zeta.nn.modules.norm_utils import PostNorm
from zeta.nn.modules.p_scan import PScan, pscan
from zeta.nn.modules.parallel_wrapper import Parallel
from zeta.nn.modules.patch_img import patch_img
from zeta.nn.modules.patch_video import patch_video
from zeta.nn.modules.perceiver_layer import PerceiverLayer
from zeta.nn.modules.poly_expert_fusion_network import MLPProjectionFusion
from zeta.nn.modules.polymorphic_activation import PolymorphicActivation
from zeta.nn.modules.polymorphic_neuron import PolymorphicNeuronLayer
from zeta.nn.modules.prenorm import PreNorm
from zeta.nn.modules.proj_then_softmax import FusedProjSoftmax
from zeta.nn.modules.pulsar import Pulsar
from zeta.nn.modules.pyro import hyper_optimize
from zeta.nn.modules.qformer import QFormer
from zeta.nn.modules.qkv_norm import qk_norm, qkv_norm
from zeta.nn.modules.quantized_layernorm import QuantizedLN
from zeta.nn.modules.recursive_block import RecursiveBlock
from zeta.nn.modules.residual import Residual
from zeta.nn.modules.resnet import ResNet
from zeta.nn.modules.rms_norm import RMSNorm
from zeta.nn.modules.rnn_nlp import RNNL
from zeta.nn.modules.shufflenet import ShuffleNet
from zeta.nn.modules.sig_lip import SigLipLoss
from zeta.nn.modules.simple_attention import simple_attention
from zeta.nn.modules.simple_feedforward import SimpleFeedForward
from zeta.nn.modules.simple_mamba import Mamba, MambaBlock
from zeta.nn.modules.simple_res_block import SimpleResBlock
from zeta.nn.modules.skipconnection import SkipConnection
from zeta.nn.modules.slerp_model_merger import SLERPModelMerger
from zeta.nn.modules.space_time_unet import (
    ContinuousPositionBias,
    Downsample,
    FeedForwardV,
    PseudoConv3d,
    ResnetBlock,
    SpaceTimeUnet,
    SpatioTemporalAttention,
    Upsample,
)
from zeta.nn.modules.spatial_transformer import SpatialTransformer
from zeta.nn.modules.ssm import SSM, selective_scan, selective_scan_seq
from zeta.nn.modules.stoch_depth import StochDepth
from zeta.nn.modules.stochastic_depth import StochasticSkipBlocK
from zeta.nn.modules.subln import SubLN
from zeta.nn.modules.super_resolution import SuperResolutionNet
from zeta.nn.modules.swiglu import SwiGLU, SwiGLUStacked
from zeta.nn.modules.time_up_sample import TimeUpSample2x
from zeta.nn.modules.to_logits import to_logits
from zeta.nn.modules.token_learner import TokenLearner
from zeta.nn.modules.top_n_gating import TopNGating
from zeta.nn.modules.triple_skip import TripleSkipBlock
from zeta.nn.modules.u_mamba import UMambaBlock
from zeta.nn.modules.unet import Unet
from zeta.nn.modules.v_layernorm import VLayerNorm
from zeta.nn.modules.v_pool import DepthWiseConv2d, Pool
from zeta.nn.modules.video_autoencoder import CausalConv3d
from zeta.nn.modules.video_diffusion_modules import (
    AttentionBasedInflationBlock,
    ConvolutionInflationBlock,
    TemporalDownsample,
    TemporalUpsample,
)
from zeta.nn.modules.video_to_tensor import video_to_tensor, video_to_tensor_vr
from zeta.nn.modules.video_to_text import video_to_text
from zeta.nn.modules.visual_expert import VisualExpert
from zeta.nn.modules.vit_denoiser import (
    VisionAttention,
    VitTransformerBlock,
    posemb_sincos_2d,
    to_patch_embedding,
)
from zeta.nn.modules.ws_conv2d import WSConv2d
from zeta.nn.modules.yolo import yolo
from zeta.nn.modules.palo_ldp import PaloLDP
from zeta.nn.modules.relu_squared import ReluSquared
from zeta.nn.modules.scale_norm import ScaleNorm
from zeta.nn.modules.mr_adapter import MRAdapter
from zeta.nn.modules.sparse_moe import (
    Top2Gating,
    NormalSparseMoE,
    HeirarchicalSparseMoE,
)
from zeta.nn.modules.return_loss_text import (
    return_loss_text,
    calc_z_loss,
    max_neg_value,
    TextTokenEmbedding,
    dropout_seq,
    transformer_generate,
)
from zeta.nn.modules.patch_linear_flatten import (
    vit_output_head,
    patch_linear_flatten,
    cls_tokens,
    video_patch_linear_flatten,
)
from zeta.nn.modules.chan_layer_norm import ChanLayerNorm

from zeta.nn.modules.query_proposal import TextHawkQueryProposal
from zeta.nn.modules.pixel_shuffling import PixelShuffleDownscale
from zeta.nn.modules.kan import KAN
from zeta.nn.modules.layer_scale import LayerScale
from zeta.nn.modules.fractoral_norm import FractoralNorm
from zeta.nn.modules.kv_cache_update import kv_cache_with_update
from zeta.nn.modules.expand import expand
from zeta.nn.modules.sig_lip_loss import SigLipSigmoidLoss
from zeta.nn.modules.sparse_token_integration import (
    SparseTokenIntegration,
    SparseChannelIntegration,
)
from zeta.nn.modules.simple_lstm import SimpleLSTM
from zeta.nn.modules.simple_rnn import SimpleRNN
from zeta.nn.modules.cope import CoPE
from zeta.nn.modules.multi_layer_key_cache import MultiLayerKeyValueAttention
from zeta.nn.modules.evlm_xattn import GatedMoECrossAttn, GatedXAttention
from zeta.nn.modules.snake_act import Snake
from zeta.nn.modules.adaptive_gating import AdaptiveGating
from zeta.nn.modules.crome_adapter import CROMEAdapter
from zeta.nn.modules.cog_vlm_two_adapter import CogVLMTwoAdapter
from zeta.nn.modules.sigmoid_attn import SigmoidAttention

# from zeta.nn.modules.img_reshape import image_reshape
# from zeta.nn.modules.flatten_features import flatten_features
# from zeta.nn.modules.scaled_sinusoidal import ScaledSinuosidalEmbedding
# from zeta.nn.modules.scale import Scale
# from zeta.nn.modules.scalenorm import ScaleNorm
# from zeta.nn.modules.simple_rmsnorm import SimpleRMSNorm
# from zeta.nn.modules.gru_gating import GRUGating
# from zeta.nn.modules.shift_tokens import ShiftTokens
# from zeta.nn.modules.swarmalator import simulate_swarmalators
# from zeta.nn.modules.transformations import image_transform
# from zeta.nn.modules.squeeze_excitation import SqueezeExcitation
# from zeta.nn.modules.clex import Clex

__all__ = [
    "CNNNew",
    "CombinedLinear",
    "ConvNet",
    "DynamicModule",
    "Exo",
    "FastTextNew",
    "FeedForwardNetwork",
    "LayerNorm",
    "l2norm",
    "Lora",
    "MBConv",
    "MLP",
    "Pulsar",
    "Residual",
    "ResNet",
    "RMSNorm",
    "RNNL",
    "ShuffleNet",
    "simple_attention",
    "SpatialTransformer",
    "SubLN",
    "SuperResolutionNet",
    "TokenLearner",
    "yolo",
    "Ether",
    "Nebula",
    "AdaptiveConv3DMod",
    "TimeUpSample2x",
    "CausalConv3d",
    "SimpleResBlock",
    "SigLipLoss",
    "SimpleFeedForward",
    "Unet",
    "VisualExpert",
    "FeedForward",
    "SkipConnection",
    "LogFF",
    "PolymorphicNeuronLayer",
    "CustomMLP",
    "PolymorphicActivation",
    "PreNorm",
    "IterativeCrossSelfAttention",
    "ConvolutionLanguageBlock",
    "H3Layer",
    "MLPMixer",
    "LeakyRELU",
    "AdaptiveLayerNorm",
    "SwiGLU",
    "SwiGLUStacked",
    "ImgPatchEmbed",
    "DenseBlock",
    "HighwayLayer",
    "MultiScaleBlock",
    "FeedbackBlock",
    "DualPathBlock",
    "RecursiveBlock",
    "PytorchGELUTanh",
    "NewGELUActivation",
    "GELUActivation",
    "FastGELUActivation",
    "QuickGELUActivation",
    "ClippedGELUActivation",
    "AccurateGELUActivation",
    "MishActivation",
    "LinearActivation",
    "LaplaceActivation",
    "ReLUSquaredActivation",
    "TripleSkipBlock",
    "DynamicRoutingBlock",
    "GatedResidualBlock",
    "StochasticSkipBlocK",
    "QuantizedLN",
    "SLERPModelMerger",
    "AverageModelMerger",
    "AdaptiveRMSNorm",
    "MambaBlock",
    "Mamba",
    "Laser",
    "FusedDenseGELUDense",
    "FusedDropoutLayerNorm",
    "Conv2DFeedforward",
    "MLPBlock",
    "MixerBlock",
    "WSConv2d",
    "StochDepth",
    "NFNStem",
    "Film",
    "DropSample",
    "SqueezeExcitation",
    "MBConvResidual",
    "video_to_tensor",
    "video_to_tensor_vr",
    "FusedProjSoftmax",
    "TopNGating",
    "MoERouter",
    "PerceiverLayer",
    "UMambaBlock",
    "audio_to_text",
    "patch_video",
    "img_to_text",
    "video_to_text",
    "hyper_optimize",
    "to_patch_embedding",
    "posemb_sincos_2d",
    "VisionAttention",
    "VitTransformerBlock",
    "VLayerNorm",
    "Parallel",
    "DepthWiseConv2d",
    "Pool",
    "MixtureOfExperts",
    "FlexiConv",
    "MMLayerNorm",
    "MMFusionFFN",
    "PostNorm",
    "PScan",
    "pscan",
    "selective_scan",
    "selective_scan_seq",
    "SSM",
    "FilmConditioning",
    "qkv_norm",
    "qk_norm",
    "FeedForwardV",
    "ContinuousPositionBias",
    "PseudoConv3d",
    "SpatioTemporalAttention",
    "ResnetBlock",
    "Downsample",
    "Upsample",
    "SpaceTimeUnet",
    "patch_img",
    "threed_to_text",
    "text_to_twod",
    "jit_dropout_add",
    "fused_dropout_add",
    "jit_bias_dropout_add",
    "fused_bias_dropout_add",
    "blockdiag_butterfly_multiply_reference",
    "BlockdiagButterflyMultiply",
    "blockdiag_weight_to_dense_weight",
    "blockdiag_multiply_reference",
    "BlockdiagMultiply",
    "fftconv_ref",
    "mul_sum",
    "Sin",
    "StructuredLinear",
    "BlockButterflyLinear",
    "BlockMLP",
    "GILLMapper",
    "add_norm",
    "to_logits",
    "CrossModalReParametrization",
    "CrossModalReparamLinear",
    "cross_modal_ffn",
    "build_cross_modal_reparam_linear",
    "change_original_linear_to_reparam",
    "reparameterize_aux_into_target_model",
    "QFormer",
    "MLPProjectionFusion",
    "NormalizationFractral",
    "image_or_video_to_time",
    "TemporalDownsample",
    "TemporalUpsample",
    "ConvolutionInflationBlock",
    "AttentionBasedInflationBlock",
    "freeze_all_layers",
    "set_module_requires_grad",
    "MultiModalEmbedding",
    "MultiInputMultiModalConcatenation",
    "SplitMultiOutput",
    "OutputHead",
    "DynamicOutputDecoder",
    "DynamicInputChannels",
    "OutputDecoders",
    "PaloLDP",
    "ReluSquared",
    "ScaleNorm",
    "MRAdapter",
    "Top2Gating",
    "NormalSparseMoE",
    "HeirarchicalSparseMoE",
    "return_loss_text",
    "calc_z_loss",
    "max_neg_value",
    "TextTokenEmbedding",
    "dropout_seq",
    "transformer_generate",
    "patch_linear_flatten",
    "vit_output_head",
    "posemb_sincos_2d",
    "ChanLayerNorm",
    "cls_tokens",
    "video_patch_linear_flatten",
    "TextHawkQueryProposal",
    "PixelShuffleDownscale",
    "KAN",
    "LayerScale",
    "FractoralNorm",
    "kv_cache_with_update",
    "expand",
    "SigLipSigmoidLoss",
    "SparseTokenIntegration",
    "SparseChannelIntegration",
    "SimpleLSTM",
    "SimpleRNN",
    "CoPE",
    "MultiLayerKeyValueAttention",
    "GatedMoECrossAttn",
    "GatedXAttention",
    "Snake",
    "AdaptiveGating",
    "CROMEAdapter",
    "CogVLMTwoAdapter",
    "SigmoidAttention",
]



import torch
from torch import einsum, nn


def exists(val):
    return val is not None


def divisible_by(a, b):
    return a % b == 0


class ScaledSinuosidalEmbedding(nn.Module):
    """
    scaled sinusoidal embedding

    Args:
        dim (int): dimension of the embedding

    Returns:
        torch.Tensor: embedding of shape (seq_len, dim)


    Usage:
        >>> embed = ScaledSinuosidalEmbedding(dim=512)
        >>> x = torch.randn(1, 1024, 512)
        >>> pos = torch.randint(0, 1024, (1, 1024))
        >>> embed(x, pos).shape
        torch.Size([1, 1024, 512])


    """

    def __init__(self, dim: int, theta=10000):
        super().__init__()
        assert divisible_by(dim, 2)

        self.scale = nn.Parameter(torch.ones(1) * dim**-0.5)

        half_dim = dim // 2
        freq_seq = torch.arange(half_dim).float() / half_dim
        inv_freq = theta**-freq_seq
        self.register_buffer("inv_freq", inv_freq, persistent=False)

    def forward(self, x, pos=None, seq_start_pos=None):
        """Forward method of scaled sinusoidal embedding"""
        seq_len, device = x.shape[1], x.device

        if not exists(pos):
            pos = torch.arange(seq_len, device=device)

        if exists(seq_start_pos):
            pos = pos - seq_start_pos[..., None]

        emb = einsum("i, j -> i j", pos, self.inv_freq)
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb * self.scale
